/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.authentication;

import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.n52.security.common.util.DateUtil;
import org.n52.security.common.util.StringUtils;
import org.n52.security.common.xml.DOMParser;
import org.n52.security.common.xml.DOMSerializer;
import org.n52.security.common.xml.DOMSerializerOptions;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * Class LicenseReference encapsulates common methods of a license reference.
 * Todo: Move This class to service.licman package because it is highly related to this domain!
 * 
 * @author Marko Reiprecht Created on 04.03.2008
 * @author Oliver Meyer
 * @version $Revision: $
 */
public class LicenseReference {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Namespace URL of the license schema.
     */
    public static final String NAMESPACE_LICENSE = "http://www.52north.org/license/0.3.2";

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(LicenseReference.class);

    private static final String NAMESPACE_SAML_ASSERTION = "urn:oasis:names:tc:SAML:2.0:assertion";

    private static final String SAML_PREFIX = "urn:opengeospatial:ows4:geodrm:";
    private static final String LICID = SAML_PREFIX + "LicenseID";

    private static final String LICMGRURL = SAML_PREFIX + "LicenseManagerURL";

    private static final String NOTBEFORE = SAML_PREFIX + "NotBefore";

    private static final String NOTONORAFTER = SAML_PREFIX + "NotOnOrAfter";

    private static final String ACTIVE = SAML_PREFIX + "Active";

    private Element m_reference;

    private Map m_attributes = new HashMap();

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance.
     * 
     * @param root
     *            the root.
     */
    public LicenseReference(final Element root) {
        if (LOG.isTraceEnabled()) {
            LOG.trace("LicenseReference.LicenseReference");
        }
        if (!NAMESPACE_LICENSE.equals(root.getNamespaceURI()) || !"LicenseReference".equals(root.getLocalName())) {
            throw new IllegalArgumentException("given element is no license reference");
        }
        m_reference = root;

        // gets all SAML Attriubtes from the license referenece, and puts them in the map,
        // for faster look ups
        m_attributes = getSAMLAttributes(m_attributes);
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the license reference.
     * 
     * @return Element.
     */
    public Element getReference() {
        return m_reference;
    }

    /**
     * Method gets the licenseId.
     * 
     * @return String.
     */
    public String getLicenseId() {
        return (String) m_attributes.get(LICID);
    }

    /**
     * Method gets the licenseManagerUrl.
     * 
     * @return String.
     */
    public String getLicenseManagerUrl() {
        return (String) m_attributes.get(LICMGRURL);
    }

    public boolean isActive() {
        String booleanValue = (String) m_attributes.get(ACTIVE);
        return booleanValue == null || "true".equalsIgnoreCase(booleanValue);
    }

    /**
     * Method validates the usage properties of this reference.
     * 
     * @throws RuntimeException
     *             if error occurs.
     */
    public void validate() throws RuntimeException {
        validate(0);
    }

    /**
     * Method validates the usage properties of this reference, with respect to time clock differences.
     * 
     * @param timeDelta
     *            allowed delta in msec.
     * @throws RuntimeException
     *             if error occurs.
     */
    public void validate(final long timeDelta) throws RuntimeException {
        // check the time constrains if present
        Calendar actualDate = Calendar.getInstance();

        String notbeforDate = (String) m_attributes.get(NOTBEFORE);
        if (notbeforDate != null) {
            Calendar notBefore = DateUtil.fromISOString(notbeforDate);
            notBefore.setTimeInMillis(notBefore.getTimeInMillis() - timeDelta);
            if (actualDate.before(notBefore)) {
                throw new IllegalStateException("License reference cannot be applied before <" + notbeforDate + ">");
            }
        }
        String notOnOrAfterDate = (String) m_attributes.get(NOTONORAFTER);
        if (notOnOrAfterDate != null) {
            Calendar notOnOrAfter = DateUtil.fromISOString(notOnOrAfterDate);
            notOnOrAfter.setTimeInMillis(notOnOrAfter.getTimeInMillis() + timeDelta);
            if (actualDate.equals(notOnOrAfter) || actualDate.after(notOnOrAfter)) {
                throw new IllegalStateException("License reference cannot be applied on or after <" + notOnOrAfterDate
                        + ">");
            }
        }
    }

    /**
     * Method gets all SAMLAttributes from the license reference as (keyString,valueString) map.
     * 
     * @param attributeMap
     *            the attributeMap (to which the attributes shall be added, can be null).
     * @return Map(String,String).
     */
    private Map getSAMLAttributes(final Map attributeMap) {
        Map resultMap = attributeMap == null ? new HashMap() : attributeMap;
        NodeList attributes = m_reference.getElementsByTagNameNS(NAMESPACE_SAML_ASSERTION, "Attribute");
        for (int i = 0, l = attributes.getLength(); i < l; ++i) {
            Element attribute = (Element) attributes.item(i);
            String name = attribute.getAttribute("Name");
            Element attributevalue =
                (Element) attribute.getElementsByTagNameNS(NAMESPACE_SAML_ASSERTION, "AttributeValue").item(0);
            String value = getText(attributevalue);
            resultMap.put(name, value);
        }
        return resultMap;
    }

    /**
     * Gets all saml attributes from the license reference.
     * 
     * @return <String,String>
     */
    public Map getSAMLAttributes() {
        return Collections.unmodifiableMap(m_attributes);
    }

    /**
     * Method getSAMLAttributeValue searches for the given attributeName in the license reference.
     * 
     * @param attributeName
     *            the attributeName.
     * @return String.
     */
    public String getSAMLAttributeValue(final String attributeName) {
        String attrvalue = (String) m_attributes.get(attributeName);
        if (attrvalue != null) {
            return attrvalue;
        }
        NodeList attributes = m_reference.getElementsByTagNameNS(NAMESPACE_SAML_ASSERTION, "Attribute");
        for (int i = 0, l = attributes.getLength(); i < l; ++i) {
            Element attribute = (Element) attributes.item(i);
            if (attributeName.equals(attribute.getAttribute("Name"))) {
                Element attributevalue =
                    (Element) attribute.getElementsByTagNameNS(NAMESPACE_SAML_ASSERTION, "AttributeValue").item(0);
                return getText(attributevalue);
            }
        }
        return null;
    }

    /**
     * Method gets the text from this element.
     * 
     * @param attributevalue
     *            the attributevalue.
     * @return String.
     */
    private String getText(final Element attributevalue) {
        NodeList childs = attributevalue.getChildNodes();
        for (int i = 0, l = childs.getLength(); i < l; ++i) {
            Node child = childs.item(i);
            if (Node.TEXT_NODE == child.getNodeType()) {
                return child.getNodeValue().trim();
            }
        }
        return null;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method creates a license reference from a stream.
     * 
     * @param stream
     *            the stream.
     * @return LicenseReference.
     */
    public static LicenseReference createFrom(final InputStream stream) {
        if (LOG.isTraceEnabled()) {
            LOG.trace("LicenseReference.createFrom(Stream)");
        }
        Document doc = DOMParser.createNew().parse(new InputSource(stream));
        return new LicenseReference(doc.getDocumentElement());
    }

    /**
     * Method creates a license reference from a reader.
     * 
     * @param reader
     *            the reader.
     * @return LicenseReference.
     */
    public static LicenseReference createFrom(final Reader reader) {
        if (LOG.isTraceEnabled()) {
            LOG.trace("LicenseReference.createFrom(Reader)");
        }
        Document doc = DOMParser.createNew().parse(new InputSource(reader));
        return new LicenseReference(doc.getDocumentElement());
    }

    /**
     * Method creates licenes ref from a string. It is a short cut for {@link #createFrom(java.io.Reader)} called with a
     * StringReader. In case string is base 64 encoded it will be decoded first.
     * 
     * @param ref
     *            the ref.
     * @return LicenseReference.
     */
    public static LicenseReference createFrom(final String ref) {
        if (LOG.isTraceEnabled()) {
            LOG.trace("LicenseReference.createFrom(String)");
        }
        return createFrom(new StringReader(StringUtils.decodeBase64(ref)));
    }

    /**
     * Method creates the license ref from document.
     * 
     * @param rootNode
     *            the rootNode.
     * @return LicenseReference.
     */
    public static LicenseReference createFrom(final Element rootNode) {
        return new LicenseReference(rootNode);
    }

    /**
     * Method gets the element reference as a string.
     * 
     * @return String.
     */
    public String referenceAsString() {
        return DOMSerializer.createNew(DOMSerializerOptions.getDefaultOptions().omitXMLDeclaration())
                .serializeToString(getReference());
    }

    /**
     * Returns license reference element as Base64 encoded string.
     * 
     * @return license reference element as Base64 encoded string
     */
    public String getReferenceAsBase64EncodedString() {
        return StringUtils.encodeBase64(referenceAsString());
    }

    /**
     * Method equals.
     * 
     * @param o
     *            the o.
     * @return boolean.
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof LicenseReference)) {
            return false;
        }

        final LicenseReference that = (LicenseReference) o;

        String licId = getLicenseId();
        String thatLicId = that.getLicenseId();

        if (licId != null ? !licId.equals(thatLicId) : thatLicId != null) {
            return false;
        }

        return true;
    }

    /**
     * Method hashCode.
     * 
     * @return int.
     */
    public int hashCode() {
        String licId = getLicenseId();
        return (licId != null ? licId.hashCode() : 0);
    }

    public String toString() {
        final StringBuffer sb = new StringBuffer();
        sb.append("org.n52.security.authentication.LicenseReference");
        sb.append("{licenseId='").append(getLicenseId()).append('\'');
        sb.append(", licenseManagerUrl='").append(getLicenseManagerUrl()).append('\'');
        sb.append('}');
        return sb.toString();
    }
}
