/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.authentication;

import java.io.IOException;
import java.io.StringReader;
import java.security.Key;
import java.security.Principal;
import java.security.cert.Certificate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.n52.security.authentication.principals.AttributePrincipal;
import org.n52.security.authentication.principals.NamedScope;
import org.n52.security.authentication.principals.RolePrincipal;
import org.n52.security.authentication.principals.Scope;
import org.n52.security.authentication.principals.UsernameIDPrincipal;
import org.n52.security.common.attributes.Attribute;
import org.n52.security.common.attributes.AttributeValue;
import org.n52.security.common.attributes.DateAttributeValue;
import org.n52.security.common.attributes.StringAttributeValue;
import org.n52.security.common.util.StringUtils;
import org.n52.security.common.xml.DOMParser;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAttribute;
import org.opensaml.SAMLAttributeStatement;
import org.opensaml.SAMLAuthenticationStatement;
import org.opensaml.SAMLException;
import org.opensaml.SAMLNameIdentifier;
import org.opensaml.SAMLResponse;
import org.opensaml.SAMLStatement;
import org.opensaml.SAMLSubject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * The SAMLTicket class basically provides a SAMLResponse <-> javax.security.auth.Subject mapping.
 */
public class SAMLTicket {

    private static Log LOG = LogFactory.getLog(SAMLTicket.class);
    public static final String SUBJECT_ATTRIBUTE_AUTH_INSTANT = "authInstant";
    public static final String SUBJECT_ATTRIBUTE_AUTH_METHOD = "authMethod";

    protected SAMLResponse m_samlResponse;
    protected SAMLAssertion m_samlAssertion;

    /**
     * @param samlResponse
     *            String representation of an "UTF-8" coded SAMLTicket.
     * @throws SAMLException
     */
    public SAMLTicket(final String samlResponseOrAssertion) throws SAMLException {
        this(DOMParser.createNew().parse(new InputSource(new StringReader(samlResponseOrAssertion))));
    }

    /**
     * Creates new instance.
     * 
     * @param pElement
     * @throws SAMLException
     */
    public SAMLTicket(Element root) throws SAMLException {
        if (org.opensaml.XML.SAMLP_NS.equals(root.getNamespaceURI()) && "Response".equals(root.getNodeName())) {
            m_samlResponse = new SAMLResponse(root);
            Iterator assertions = m_samlResponse.getAssertions();
            if (!assertions.hasNext()) {
                throw new IllegalArgumentException("<samlResponse> contains no saml assertion!");
            }
            m_samlAssertion = (SAMLAssertion) assertions.next();
            if (assertions.hasNext()) {
                LOG.warn("SAMLResponse contains more than one assertion, only the first is used!");
            }
        } else if (org.opensaml.XML.SAML_NS.equals(root.getNamespaceURI()) && "Assertion".equals(root.getNodeName())) {
            m_samlAssertion = new SAMLAssertion(root);
        } else {
            throw new IllegalArgumentException("element <" + root.getNodeName() + "> with namespace <"
                    + root.getNamespaceURI()
                    + "> is not supported.");
        }
    }

    /**
     * @param pSAMLResponse
     * @throws SAMLException
     */
    public SAMLTicket(Document samlResponseOrAssertion) throws SAMLException {
        this(samlResponseOrAssertion.getDocumentElement());
    }

    /**
     * @param pSAMLResponse
     * @throws SAMLException
     */
    public SAMLTicket(final SAMLResponse samlResponse) throws SAMLException {
        if (samlResponse == null) {
            throw new IllegalArgumentException("<samlResponse> must not null");
        }
        m_samlResponse = samlResponse;
        Iterator assertions = samlResponse.getAssertions();
        if (!assertions.hasNext()) {
            throw new IllegalArgumentException("<samlResponse> contains no saml assertion!");
        }
        m_samlAssertion = (SAMLAssertion) assertions.next();
        if (assertions.hasNext()) {
            LOG.warn("SAMLResponse contains more than one assertion, only the first is used!");
        }
    }

    public SAMLTicket(final SAMLAssertion samlAssertion) throws SAMLException {
        if (samlAssertion == null) {
            throw new IllegalArgumentException("<samlAssertion> must not null");
        }
        m_samlAssertion = samlAssertion;
    }

    /**
     * Verifies the signature of this SAML document.
     * 
     * @throws SAMLException
     *             if the signature could not be verified
     */
    public void verify() throws SAMLException {
        if (m_samlResponse != null) {
            m_samlResponse.verify();
        }
        if (m_samlAssertion != null) {
            m_samlAssertion.verify();
        }
    }

    public void verify(final Certificate cert) throws SAMLException {
        if (cert == null) {
            verify();
        } else {
            boolean samlResponseSigned = false;
            if (m_samlResponse != null && m_samlResponse.isSigned()) {
                m_samlResponse.verify(cert);
                samlResponseSigned = true;
            }
            boolean samlAssertionSigned = false;
            if (m_samlAssertion != null && m_samlAssertion.isSigned()) {
                m_samlAssertion.verify(cert);
                samlAssertionSigned = true;
            }
            if (!samlResponseSigned && !samlAssertionSigned) {
                throw new SAMLException("SAML Response and/or Assertion are not signed!");
            }
        }
    }

    /**
     * check if any assertion is expired, without any respect to clock differences.
     * 
     * @return if ticket is expired.
     */
    public boolean isExpired() {
        return isExpired(0);
    }

    /**
     * check if any assertion is expired, in relation to an allowed delta.
     * 
     * @param allowedDelta
     *            a delta in msec.
     * @return if ticket is expired.
     */
    public boolean isExpired(final long allowedDelta) {
        final long lNow = System.currentTimeMillis();
        SAMLAssertion tAssert = m_samlAssertion;
        final Date notBefore = tAssert.getNotBefore();
        if (notBefore != null && (notBefore.getTime() - allowedDelta) > lNow) {
            return true;
        }
        final Date notOnOrAfter = tAssert.getNotOnOrAfter();
        if (notOnOrAfter != null && (notOnOrAfter.getTime() + allowedDelta) <= lNow) {
            return true;
        }
        return false;
    }

    public Date getNotOnOrAfter() {
        return m_samlAssertion.getNotOnOrAfter();
    }

    public Date getNotBefore() {
        return m_samlAssertion.getNotBefore();
    }

    /**
     * @return
     */
    public String asString() {
        return m_samlResponse != null ? m_samlResponse.toString() : m_samlAssertion.toString();
    }

    /**
     * @return
     */
    public String asBase64String() {
        return StringUtils.encodeBase64(m_samlResponse != null ? m_samlResponse.toString() : m_samlAssertion.toString());
    }

    /**
     * @return
     */
    public String getAssertionAsBase64String() {
        return StringUtils.encodeBase64(m_samlAssertion.toString());
    }

    public Element getAssertionAsDOM() {
        // ensure to return copy
        return DOMParser.createNew().parse(new InputSource(new StringReader(m_samlAssertion.toString()))).getDocumentElement();
    }

    /**
     * @return
     * @deprecated user {@link #asSubject(String)}
     */
    public Subject asSubject() {
        Subject lSubject = new Subject();
        updateSubject(lSubject);
        return lSubject;
    }

    public Subject asSubject(String roleAttributeId) {
        Subject lSubject = new Subject();
        updateSubject(lSubject, roleAttributeId);
        return lSubject;
    }

    /**
     * @param pSubject
     * @deprecated use {@link #updateSubject(Subject, String)}
     */
    public void updateSubject(Subject pSubject) {
        updateSubject(pSubject, "");
    }

    public void updateSubject(Subject pSubject, String roleAttributeId) {
        if (pSubject == null) {
            throw new IllegalArgumentException("Subject must not be null.");
        }
        if (roleAttributeId == null) {
            roleAttributeId = "urn:conterra:names:sdi-suite:policy:attribute:role";
        }
        pSubject.getPrincipals().addAll(retrievePrincipals(roleAttributeId));
    }

    /**
     * @return
     */
    private Set retrievePrincipals(String roleAttributeId) {
        HashSet lPrincipals = new HashSet();
        SAMLAssertion samlFirstAssertion = m_samlAssertion;

        // Iterate through statements and create appropriate principals
        Iterator itrSamlStatements = samlFirstAssertion.getStatements();
        while (itrSamlStatements.hasNext()) {
            SAMLStatement tSamlStatement = (SAMLStatement) itrSamlStatements.next();

            // evaluate AuthenticationStatement
            if (tSamlStatement instanceof SAMLAuthenticationStatement) {
                SAMLAuthenticationStatement samlAuthnStatement = (SAMLAuthenticationStatement) tSamlStatement;
                SAMLSubject samlSubject = samlAuthnStatement.getSubject();
                String userName = samlSubject.getNameIdentifier().getName();
                String userNameQualifier = samlSubject.getNameIdentifier().getNameQualifier();
                lPrincipals.add(new UsernameIDPrincipal(userName, userNameQualifier == null ? Scope.GLOBAL
                        : new NamedScope(userNameQualifier)));
                Date authInstant = samlAuthnStatement.getAuthInstant();
                if (authInstant != null) {
                    lPrincipals.add(new AttributePrincipal(new Attribute(SUBJECT_ATTRIBUTE_AUTH_INSTANT,
                            new DateAttributeValue(authInstant))));
                }
                String method = samlAuthnStatement.getAuthMethod();
                if (method != null) {
                    lPrincipals.add(new AttributePrincipal(new Attribute(SUBJECT_ATTRIBUTE_AUTH_METHOD,
                            new StringAttributeValue(method))));
                }
            }

            // evaluate AttributeStatements
            if (tSamlStatement instanceof SAMLAttributeStatement) {
                SAMLAttributeStatement samlAttrStatement = (SAMLAttributeStatement) tSamlStatement;

                Iterator itrSamlAttributes = samlAttrStatement.getAttributes();
                while (itrSamlAttributes.hasNext()) {
                    SAMLAttribute tAtt = (SAMLAttribute) itrSamlAttributes.next();
                    Iterator itrSamlAttrValues = tAtt.getValues();
                    while (itrSamlAttrValues.hasNext()) {
                        String name = tAtt.getName();
                        String value = (String) itrSamlAttrValues.next();
                        if (roleAttributeId.length() > 0 && roleAttributeId.equals(name)) {
                            lPrincipals.add(new RolePrincipal(value));
                        } else {
                            // TODO: support the different xml types!
                            lPrincipals.add(new AttributePrincipal(tAtt.getName(), value));
                        }
                    }
                }
            }
        }
        return lPrincipals;

    }

    /**
     * @throws IOException
     * @throws SAMLException
     * @see #createSAMLResponseFromPrincipals(AuthenticationContext, String, String, String, int, String, Key,
     *      Certificate, boolean, Set)
     */
    public static String createSAMLResponseFromPrincipals(AuthenticationContext authContext,
                                                          String samlRoleAttributeName,
                                                          String samlIssuerName,
                                                          String target,
                                                          int samlAssertionTimeOut,
                                                          String authMethod,
                                                          Key privatekey,
                                                          Certificate cert,
                                                          boolean base64) throws SAMLException, IOException {
        return createSAMLResponseFromPrincipals(authContext, samlRoleAttributeName, null, samlIssuerName, target,
                samlAssertionTimeOut, authMethod, privatekey, cert, base64, true, null);
    }

    /**
     * Creates a SAML ticket, i.e. a SAMLResponse, based on a set of {@link Principal} instances that are part of a
     * {@link Subject} stored in an {@link AuthenticationContext} instance.
     * 
     * @param authContext
     *            context of an authentication process that took place in the past. The {@link Subject} provided by
     *            {@link AuthenticationContext#getSubject()} must contain exactly one {@link UsernameIDPrincipal}. Every
     *            {@link AttributePrincipal} of the Subject is mapped to a SAMLAttributeSatement. Every
     *            {@link RolePrincipal} is mapped to a special SAMLAttributeStatement that will be interpreted as roles
     *            during authorization.
     * @param pRoleAttributeName
     *            Specifies the name of the AttributeStatements used to store role information, e.g.
     *            <code>urn:conterra:names:sdi-suite:policy:attribute:role</code>
     * @param pIssuer
     *            The name that will appear as issuer of the SAMLAssertion.
     * @param pTarget
     *            The name of the consumer of the SAMLAssertion.
     * @param pTimeout
     *            The period of validity of the SAMLAssertion in seconds.
     * @param pAuthentMethod
     *            The authentication method that was used to identify the subject, e.g. <code>#password</code>. Go to
     *            the SAML specification part called <i>Assertions and Protocol for the OASIS Security Assertion Markup
     *            Language (SAML) V1.1</i> to find a list of authentication method IDs
     * @param pPrivKey
     *            The private key that will be used to sign the SAML Response.
     * @param pCert
     *            The certificate that belongs to the private key. It will be embedded into the SAMLResponse as part of
     *            the signature. If this is <code>null</code>, the ticket is not signed.
     * @param pBase64
     *            If <code>true</code> the SAML ticket will be Base64-encoded, it won't be encoded otherwise.
     * @return The ticket XML as a String.
     * @throws SAMLException
     * @throws IOException
     */
    public static String createSAMLResponseFromPrincipals(AuthenticationContext authContext,
                                                          String pRoleAttributeName,
                                                          String samlAttributeNamespace,
                                                          String pIssuer,
                                                          String pTarget,
                                                          int pTimeout,
                                                          String pAuthentMethod,
                                                          Key pPrivKey,
                                                          Certificate pCert,
                                                          boolean pBase64,
                                                          boolean pSignResponseElement,
                                                          Set pSamlAttributeNames) throws SAMLException, IOException {
        SAMLTicket ticket =
            createSAMLTicketFromPrincipals(authContext, pRoleAttributeName, samlAttributeNamespace, pIssuer, pTarget,
                pTimeout, pAuthentMethod, pPrivKey, pCert, pBase64, pSignResponseElement, pSamlAttributeNames);

        // use normal utilities
        return pBase64 ? ticket.asBase64String() : ticket.asString();

    }

    public static SAMLTicket createSAMLTicketFromPrincipals(AuthenticationContext authContext,
                                                            String pRoleAttributeName,
                                                            String samlAttributeNamespace,
                                                            String pIssuer,
                                                            String pTarget,
                                                            int pTimeout,
                                                            String pAuthentMethod,
                                                            Key pPrivKey,
                                                            Certificate pCert,
                                                            boolean pBase64,
                                                            boolean pSignResponseElement,
                                                            Set pSamlAttributeNames) throws SAMLException, IOException {
        Subject subject = authContext.getSubject();
        // expect at least on userid principal instance
        UsernameIDPrincipal userId =
            (UsernameIDPrincipal) subject.getPrincipals(UsernameIDPrincipal.class).iterator().next();
        String lName = userId.getName();
        String lNameQualifier = userId.getScope().getName();

        SAMLAssertion assertion;
        SAMLStatement lAuthentStatement = null;
        SAMLSubject samlSubject;
        String subjectIP = null;
        String subjectDNS = null;
        java.util.Date authInstant = authContext.getAuthenticationTime();
        AttributeValue authInstantVal = SubjectUtil.getComplexAttributeValue(subject, SUBJECT_ATTRIBUTE_AUTH_INSTANT);
        if (authInstantVal != null && ((DateAttributeValue) authInstantVal).getValue() != null) {
            // overwrite value within the authContext with the original one of the subject
            authInstant = ((DateAttributeValue) authInstantVal).getValue();
        }
        AttributeValue authMethodVal = SubjectUtil.getComplexAttributeValue(subject, SUBJECT_ATTRIBUTE_AUTH_METHOD);
        if (authMethodVal != null && ((StringAttributeValue) authMethodVal).getValue() != null) {
            // overwrite with original auth method
            pAuthentMethod = ((StringAttributeValue) authMethodVal).getValue();
        }

        SAMLResponse response;

        Calendar cal = Calendar.getInstance();
        Calendar cal0 = (Calendar) cal.clone();
        cal.add(Calendar.SECOND, pTimeout);

        SAMLNameIdentifier identifier =
            new SAMLNameIdentifier(lName, lNameQualifier, "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified");

        // Collection confMethods = new ArrayList(1);
        // confMethods.add("confirmationMethods");
        samlSubject = new SAMLSubject(identifier, null, null, null);

        Collection statements = new ArrayList();

        try {
            lAuthentStatement =
                new SAMLAuthenticationStatement((SAMLSubject) samlSubject.clone(), pAuthentMethod, authInstant,
                    subjectIP, subjectDNS, null);

            statements.add(lAuthentStatement);

            Map attributeValuesMapping = getAttributesFromSubject(subject, pRoleAttributeName, pSamlAttributeNames);
            Collection samlAtts = new ArrayList();
            Iterator lAttTableItr = attributeValuesMapping.entrySet().iterator();
            while (lAttTableItr.hasNext()) {
                Map.Entry tEntry = (Map.Entry) lAttTableItr.next();
                // TODO: currently only string attributes are produced
                SAMLAttribute tAtt =
                    new SAMLAttribute((String) tEntry.getKey(), samlAttributeNamespace == null ? "def"
                        : samlAttributeNamespace, new QName(org.opensaml.XML.XSD_NS, "string"), 0,
                        (Collection) tEntry.getValue());
                samlAtts.add(tAtt);
            }
            if (!samlAtts.isEmpty()) {
                SAMLAttributeStatement tAttStatement =
                    new SAMLAttributeStatement((SAMLSubject) samlSubject.clone(), samlAtts);
                statements.add(tAttStatement);
            }
        } catch (CloneNotSupportedException e) {
            LOG.error("Error during SAMLAuthentication Statement creation.", e);
        }

        assertion = new SAMLAssertion(pIssuer, cal0.getTime(), cal.getTime(), null, null, statements);
        Collection assertions = new ArrayList(1);
        assertions.add(assertion);
        response = new SAMLResponse(null, pTarget, assertions, null);
        // ensure canonicalization
        String resp = response.toString();
        if (LOG.isDebugEnabled()) {
            LOG.debug("Generated SAMLResponse (yet unsigned):\n" + resp);
        }

        if (pCert != null) {
            Collection certList = new ArrayList(1);
            certList.add(pCert);
            assertion.sign("http://www.w3.org/2000/09/xmldsig#rsa-sha1", pPrivKey, certList);
            if (pSignResponseElement) {
                response.sign("http://www.w3.org/2000/09/xmldsig#rsa-sha1", pPrivKey, certList);
            }
        }
        return new SAMLTicket(response);
    }

    private static Map getAttributesFromSubject(final Subject subject,
                                                final String roleAttName,
                                                final Set pSamlAttributeNames) {
        Map attributes = new HashMap();
        Iterator attributeprincipals = subject.getPrincipals(AttributePrincipal.class).iterator();
        while (attributeprincipals.hasNext()) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) attributeprincipals.next();
            String attName = attributePrincipal.getName();
            if ((pSamlAttributeNames != null && !pSamlAttributeNames.isEmpty() && !pSamlAttributeNames.contains(attName))
                    || SUBJECT_ATTRIBUTE_AUTH_INSTANT.equals(attName) || SUBJECT_ATTRIBUTE_AUTH_METHOD.equals(attName)) {
                continue;
            }
            Collection attValues = (Collection) attributes.get(attName);
            if (attValues == null) {
                attValues = new LinkedList();
                attributes.put(attName, attValues);
            }
            // todo: how to support complex attribute mappings with a static method
            String attValue = attributePrincipal.getValue();
            attValues.add(attValue);
        }

        if (pSamlAttributeNames != null && !pSamlAttributeNames.isEmpty() && !pSamlAttributeNames.contains(roleAttName)) {
            return attributes;
        }
        // skipped if not configured in saml attribute names.
        Iterator roleprincipals = subject.getPrincipals(RolePrincipal.class).iterator();
        Collection roles = new LinkedList();
        while (roleprincipals.hasNext()) {
            RolePrincipal rolePrincipal = (RolePrincipal) roleprincipals.next();
            // todo use scope ?
            roles.add(rolePrincipal.getName());
        }
        if (!roles.isEmpty()) {
            attributes.put(roleAttName, roles);
        }
        // todo explicite support for groups and/or other principals?
        return attributes;
    }
}
