/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.authentication;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.security.auth.Subject;
import org.apache.log4j.Logger;
import org.n52.security.authentication.principals.AttributePrincipal;
import org.n52.security.authentication.principals.RolePrincipal;
import org.n52.security.authentication.principals.UsernameIDPrincipal;
import org.n52.security.common.attributes.Attribute;
import org.n52.security.common.attributes.AttributeValue;

/**
 * Retrieves user attributes from a subject that is made up of JAAS principals and credentials.
 * The subject is assumed to be "validated/authenticated" by some means outside the scope of this class.
 * 
 * @author jdr
 */
public class SubjectUtil {

    private static final Logger LOG = Logger.getLogger(SubjectUtil.class);

    private SubjectUtil() {
    }

    /**
     * Retrieves any SAML token available in the given subject.
     * 
     * @param subject
     *            subject that is searched for SAML Response credentials in its public credentials .
     * @return the first {@link SAMLCredential} found in the subject's public credentials.
     */
    public static SAMLCredential getSamlToken(final Subject subject) {
        Set samlCredentials;
        samlCredentials = subject.getPublicCredentials(SAMLCredential.class);
        if (samlCredentials.size() > 1) {
            LOG.warn("Found " + samlCredentials.size() + " SAMLCredentials for subject <" + subject
                    + "> although only 1 expected. If more than one is found, one will be selected arbitrarily.");
        }
        for (Iterator itr = samlCredentials.iterator(); itr.hasNext();) {
            return (SAMLCredential) itr.next();

        }
        return null;
    }

    /**
     * Resolves the first license reference from the subject.
     * 
     * @param subject
     * @return
     */
    public static LicenseReference getLicenseReference(final Subject subject) {
        Set principals = subject.getPrincipals(LicensePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            LicensePrincipal licPrincipal = (LicensePrincipal) itr.next();
            return licPrincipal.getLicenseReference();
        }
        return null;
    }

    /**
     * Retrieves the value of the first {@link AttributePrincipal} with a given name.
     * 
     * @param subject
     *            subject containing the attribute information
     * @param attributeName
     *            the name of the attribute that contains the information.
     * @return the String value of the first {@link AttributePrincipal} of the subject
     *         where {@link AttributePrincipal#getName()} equals <code>attributeName</code>, <code>null</code> if no
     *         such exists.
     */
    public static String getSingleAttributeValue(final Subject subject, final String attributeName) {
        Set principals = subject.getPrincipals(AttributePrincipal.class);

        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) itr.next();
            if (attributePrincipal.getName().equals(attributeName)) {
                return attributePrincipal.getValue();
            }
        }
        return null;
    }

    /**
     * Retrieves the {@link AttributeValue} of the first {@link AttributePrincipal} with a given name.
     * 
     * @param subject
     *            subject containing the attribute information
     * @param attributeName
     *            the name of the attribute that contains the information.
     * @return the String value of the first {@link AttributePrincipal} of the subject
     *         where {@link AttributePrincipal#getName()} equals <code>attributeName</code>, <code>null</code> if no
     *         such exists.
     */
    public static AttributeValue getComplexAttributeValue(final Subject subject, final String attributeName) {
        Set principals = subject.getPrincipals(AttributePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) itr.next();
            if (attributePrincipal.getName().equals(attributeName)) {
                return attributePrincipal.getAttributeValue();
            }
        }
        return null;
    }

    /**
     * Creates a map of all attributes with (the last) values as string.
     * 
     * @return Map<String,String>
     */
    public static Map getSimpleAttributeValues(final Subject subject) {
        Map attributes = new HashMap();
        Set principals = subject.getPrincipals(AttributePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) itr.next();
            attributes.put(attributePrincipal.getName(), attributePrincipal.getValue());
        }
        return attributes;
    }

    /**
     * Creates a map of all attributes with (the last) values as {@link AttributeValue}.
     * 
     * @return Map<String,{@link AttributeValue}>
     */
    public static Map getComplexAttributeValues(final Subject subject) {
        Map attributes = new HashMap();
        Set principals = subject.getPrincipals(AttributePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) itr.next();
            attributes.put(attributePrincipal.getName(), attributePrincipal.getAttributeValue());
        }
        return attributes;
    }

    /**
     * Creates a list of {@link Attribute}s within the subject.
     * 
     * @param subject
     * @return
     */
    public static List getAttributes(final Subject subject) {
        List attributes = new ArrayList();
        Set principals = subject.getPrincipals(AttributePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            AttributePrincipal attributePrincipal = (AttributePrincipal) itr.next();
            attributes.add(attributePrincipal.getAttribute());
        }
        return attributes;
    }

    /**
     * Retrieves all RolePrincipal names form the subject.
     * 
     * @param subject
     *            the current subject.
     * @return
     */
    public static String[] getRolesAsArray(final Subject subject) {
        return (String[]) getRoles(subject).toArray(new String[0]);
    }

    /**
     * Retrieves all RolePrincipal names form the subject.
     * 
     * @param subject
     *            the current subject.
     * @return
     */
    public static Set getRoles(final Subject subject) {
        Set roles = new HashSet();
        Set principals = subject.getPrincipals(RolePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            RolePrincipal role = (RolePrincipal) itr.next();
            roles.add(role.getName());
        }
        return roles;
    }

    /**
     * Responsible for checking if a user has a given role.
     * 
     * @param subject
     *            the current subject.
     * @param role
     *            a role name.
     * @return
     */
    public static boolean isInRole(final Subject subject, final String role) {
        Set principals = subject.getPrincipals(RolePrincipal.class);
        for (Iterator itr = principals.iterator(); itr.hasNext();) {
            if (((RolePrincipal) itr.next()).getName().equals(role)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Retrieves the username of the subject.
     * 
     * @param subject
     *            containing the username information.
     * @return the name of the first {@link UsernameIDPrincipal} found, <code>null</code> otherwise.
     */
    public static String getUsername(final Subject subject) {
        Set usernamePrincipals = subject.getPrincipals(UsernameIDPrincipal.class);
        if (usernamePrincipals.size() != 1) {
            LOG.warn("Found " + usernamePrincipals.size() + " UsernameIDPrincipals for subject <" + subject
                    + "> although only 1 expected. If more than one is found, one will be selected arbitrarily.");
        }
        for (Iterator itr = usernamePrincipals.iterator(); itr.hasNext();) {
            UsernameIDPrincipal usernameIDPrincipal = (UsernameIDPrincipal) itr.next();
            return usernameIDPrincipal.getName();
        }
        return null;
    }
}
