/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.authentication.loginmodule;

import java.io.Serializable;
import java.util.Map;

/**
 * Class SharedState represents the shared state between login modules.
 * <p/>
 * This class is introduced to encapsulate the access to common shared states
 * and therefor to minimize programming errors.
 * <br/>
 * This class is created in the initialize method of
 * {@link org.n52.security.authentication.loginmodule.AbstractLoginModule}
 * and provided via a getter method to sub classes.
 *
 * @author Marko Reiprecht
 *         Created on 15.09.2007
 * @version $Revision: $
 */
public class SharedState implements Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Field KEY_SHARED_PASSWORD identifies a password in the shared map.
     * Please use {@link #getLoginPassword()} and {@link #setLoginPassword(char[])} to access the state information.
     *
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#login">
     *      Description of the login method in the LoginModule Developer's Guide
     *      </a>
     */
    public static final String KEY_SHARED_PASSWORD = "javax.security.auth.login.password";
    /**
     * Field KEY_SHARED_NAME  identifies a name (user/login) in the shared map.
     * Please use {@link #getLoginName()} and {@link #setLoginName(String)} to access the state information.
     *
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#login">
     *      Description of the login method in the LoginModule Developer's Guide
     *      </a>
     */
    public static final String KEY_SHARED_NAME = "javax.security.auth.login.name";
    /**
     * The serialVersionUID.
     *
     * @see java.io.Serializable
     */
    private static final long serialVersionUID = 4995300418444498284L;

    /**
     * Field m_state thats the wrapped shared state map.
     */
    private Map m_state;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor SharedState creates a new SharedState instance.
     *
     * @param state of type Map
     */
    public SharedState(final Map state) {
        assert state != null;
        m_state = state;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method clears all state information.
     */
    public void clear() {
        m_state.clear();
    }

    /**
     * Method gets the loginName of this SharedState object.
     *
     * @return String the login name, can be null if not set.
     * @see #KEY_SHARED_NAME
     */
    public String getLoginName() {
        return (String) get(KEY_SHARED_NAME);
    }

    /**
     * Method gets the state information for the key.
     *
     * @param key of type String identifying a state information.
     * @return Object or null if (!contains(key))
     */
    public Object get(final String key) {
        return m_state.get(key);
    }

    /**
     * Method gets the loginPassword of this SharedState object.
     *
     * @return String the login password, can be null if not set.
     * @see #KEY_SHARED_PASSWORD
     */
    public char[] getLoginPassword() {
        return (char[]) get(KEY_SHARED_PASSWORD);
    }

    /**
     * Methods checks for an empty shared state.
     *
     * @return boolean true if no state information is available.
     */
    public boolean isEmpty() {
        return m_state.isEmpty();
    }

    /**
     * Method checks if a shared login name is set.
     *
     * @return boolean.
     * @see #KEY_SHARED_NAME
     */
    public boolean isLoginNameSet() {
        return contains(KEY_SHARED_NAME);
    }

    /**
     * Method checks if the state information contains the key.
     *
     * @param key of type String identifying a state information.
     * @return boolean true if key in state.
     */
    public boolean contains(final String key) {
        return m_state.containsKey(key);
    }

    /**
     * Method checks if a shared login password is set.
     *
     * @return boolean.
     * @see #KEY_SHARED_PASSWORD
     */
    public boolean isLoginPasswordSet() {
        return contains(KEY_SHARED_PASSWORD);
    }

    /**
     * Method removes the login name from the shared state.
     */
    public void removeLoginName() {
        remove(KEY_SHARED_NAME);
    }

    /**
     * Method remove state information registered for the key.
     *
     * @param key of type String identifying a state information.
     * @return Object the removed state information.
     */
    public Object remove(final String key) {
        return m_state.remove(key);
    }

    /**
     * Method removes the login password from the shared state.
     */
    public void removeLoginPassword() {
        remove(KEY_SHARED_PASSWORD);
    }

    /**
     * Method sets the loginName of this SharedState object.
     *
     * @param name the loginName.
     * @see #KEY_SHARED_NAME
     */
    public void setLoginName(final String name) {
        put(KEY_SHARED_NAME, name);
    }

    /**
     * Method registers a state information for the key.
     * State information stored with the same key are overwritten.
     *
     * @param key   of type String identifying the state information.
     * @param value of type Object a new state information.
     */
    public void put(final String key, final Object value) {
        m_state.put(key, value);
    }

    /**
     * Method sets the loginPassword of this SharedState object.
     *
     * @param password the shared password.
     * @see #KEY_SHARED_PASSWORD
     */
    public void setLoginPassword(final char[] password) {
        put(KEY_SHARED_PASSWORD, password);
    }
}
