/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.07.2008
 * Modified: $Date $
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.crypto;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.Principal;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;

/**
 * Class KeystoreKeyPairProvider implements a KeyPairProvider working on a java keystore.
 *
 * @author Marko Reiprecht
 *         Created on 18.07.2008
 * @version $Revision: $
 */
public class KeystoreKeyPairProvider implements KeyPairProvider {

    // ------------------------------ FIELDS ------------------------------

    private KeyStore m_keyStore;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance from a pre loaded keystore.
     *
     * @param keyStore the keyStore.
     */
    public KeystoreKeyPairProvider(final KeyStore keyStore) {
        m_keyStore = keyStore;
    }

    /**
     * Constructor creates a new instance from a inputstream.
     *
     * @param keyStoreType the keyStoreType.
     * @param instream     the instream.
     * @param password     the password.
     */
    public KeystoreKeyPairProvider(final String keyStoreType, final InputStream instream, final char[] password) {
        try {
            KeyStore keyStore = KeyStore.getInstance(keyStoreType);
            keyStore.load(instream, password);
            m_keyStore = keyStore;
        } catch (KeyStoreException e) {
            throw new KeyPairResolvingException("keystore type <" + keyStoreType + "> not provided", e);
        } catch (IOException e) {
            throw new KeyPairResolvingException("keystore type could not loaded", e);
        } catch (NoSuchAlgorithmException e) {
            throw new KeyPairResolvingException("certificate or key encryption algorithm not supported", e);
        } catch (CertificateException e) {
            throw new KeyPairResolvingException("certificate could not be loaded", e);
        }
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the keyStore.
     *
     * @return KeyStore.
     */
    public KeyStore getKeyStore() {
        return m_keyStore;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface KeyPairProvider ---------------------

    /**
     * @see KeyPairProvider#resolveByAlias(String, char[])
     */
    public KeyPair resolveByAlias(final String alias, final char[] password) throws KeyPairResolvingException {
        try {
            DefaultKeyPair keyPair = new DefaultKeyPair();
            keyPair.setAlias(alias == null ? "" : alias);
            KeyStore keystore = getKeyStore();
            if (password != null && keystore.isKeyEntry(alias)) {
                try {
                    Key key = keystore.getKey(alias, password);
                    keyPair.setPrivateKey(key);
                } catch (NoSuchAlgorithmException e) {
                    throw new KeyPairResolvingException("private key not loadable", e);
                } catch (UnrecoverableKeyException e) {
                    throw new KeyPairResolvingException("private key not recoverable (wrong password)", e);
                }
            }
            if (keystore.isKeyEntry(alias) || keystore.isCertificateEntry(alias)) {
                Certificate cert = keystore.getCertificate(alias);
                keyPair.setCertificate(cert);
                Certificate[] chain = keystore.getCertificateChain(alias);
                keyPair.setCertificateChain(chain);
            }
            if (!keyPair.isPrivateKeySet() && !keyPair.isPublicKeySet()) {
                throw new KeyPairResolvingException("keyPair with alias <" + alias + "> can't be found");
            }
            return keyPair;
        } catch (KeyStoreException e) {
            throw new KeyPairResolvingException("keystore not accessible", e);
        }
    }

    /**
     * @see KeyPairProvider#resolveBySerialIssuer(String, BigInteger, char[])
     */
    public KeyPair resolveBySerialIssuer(final String issuerDN, final BigInteger serial, final char[] password)
            throws KeyPairResolvingException {
        KeyStore keyStore = getKeyStore();
        try {
            Enumeration aliases = keyStore.aliases();
            while (aliases.hasMoreElements()) {
                String alias = (String) aliases.nextElement();
                if (keyStore.isCertificateEntry(alias)) {
                    Certificate testCert = keyStore.getCertificate(alias);
                    if (testCert instanceof X509Certificate
                        && equalIssuerDN(((X509Certificate) testCert).getIssuerDN(), issuerDN)
                        && ((X509Certificate) testCert).getSerialNumber().equals(serial)) {
                        return resolveByAlias(alias, password);
                    }
                }
            }
            throw new KeyPairResolvingException(
                    "KeyPair for IssuerDN <" + issuerDN + "> and serial <" + serial + "> not found");
        } catch (KeyStoreException e) {
            throw new KeyPairResolvingException("keystore not accessible", e);
        }
    }

    /**
     * @see KeyPairProvider#resolveByCertificate(Certificate, char[])
     */
    public KeyPair resolveByCertificate(final Certificate cert, final char[] password)
            throws KeyPairResolvingException {
        try {
            // todo : check semantic
            String alias = getKeyStore().getCertificateAlias(cert);
            return resolveByAlias(alias, password);
        } catch (KeyStoreException e) {
            throw new KeyPairResolvingException("certificate not found in keystore", e);
        }
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method equals the issuer dn from the principal with the given issuerDN.
     *
     * @param issuer   the issuer.
     * @param issuerDN the issuerDN.
     * @return boolean.
     */
    private boolean equalIssuerDN(final Principal issuer, final String issuerDN) {
        X509NameTokenizer certIssuer = new X509NameTokenizer(issuer.getName());
        X509NameTokenizer toTest = new X509NameTokenizer(issuerDN);
        while (certIssuer.hasMoreTokens()) {
            if (!toTest.hasMoreTokens() || !certIssuer.nextToken().equals(toTest.nextToken())) {
                return false;
            }
        }
        return !toTest.hasMoreTokens();
    }
}
