package org.n52.security.common.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.log4j.Logger;

/**
 * The PathWildCardMatch function works like the ant path filter expressions.
 * <ul>
 * <li>'**' means all child paths including '/'</li>
 * <li>'*' means only one child path excluding '/'</li>
 * </ul>
 * <p/>
 * Example: /path/'*'/file matches /path/<childpath>/file<br/>
 * /path/'**'/file matches /path/<childpath>/file or /path/<childpath>/<nextpath>/file<br/>
 * <p/>
 * The match function works case insensitive.
 * 
 * @author Marko Reiprecht
 */
public class PathWildcardMatcher {
    private static final Logger LOG = Logger.getLogger(PathWildcardMatcher.class);
    private static final Pattern STARS_REGEX = Pattern.compile("(?:(/?\\*\\*/?)|(/?\\*/?))");

    private Pattern m_wildcardPattern;

    public PathWildcardMatcher(final String wildCardString, final boolean caseInsensitive) {
        m_wildcardPattern = buildPattern(wildCardString, caseInsensitive);
    }

    /**
     * @return the wildcardPattern
     */
    public Pattern getWildcardPattern() {
        return m_wildcardPattern;
    }

    public boolean matches(final String testString) {
        Pattern p = getWildcardPattern();
        boolean result = p.matcher(testString != null ? testString : "").matches();
        if (LOG.isDebugEnabled()) {
            LOG.debug("'" + p.pattern() + "' matches '" + testString + "' = " + result);
        }
        return result;
    }

    protected Pattern buildPattern(final String wildCardString, final boolean caseInsensitive) {
        String wildCard = wildCardString != null ? wildCardString : "";
        Matcher m = STARS_REGEX.matcher(wildCard);
        StringBuffer newRegex = new StringBuffer(wildCard.length() * 2);
        while (m.find()) {
            newRegex.append("\\Q");
            m.appendReplacement(newRegex, "");
            newRegex.append("\\E");
            if (m.group(1) != null) {
                // matches '/**/'
                if (m.group(1).startsWith("/")) {
                    newRegex.append("/");
                }
                newRegex.append(".*");
            } else if (m.group(2) != null) {
                // matches '/*/'
                if (m.group(2).startsWith("/")) {
                    newRegex.append("[/\\\\]+[^/\\\\]*[/\\\\]*");
                } else if (m.group(2).endsWith("/")) {
                    newRegex.append("[^/\\\\]*[/\\\\]*");
                } else {
                    newRegex.append("[^/\\\\]*");
                }
            }
        }
        newRegex.append("\\Q");
        m.appendTail(newRegex);
        newRegex.append("\\E");
        if (caseInsensitive) {
            return Pattern.compile(newRegex.toString(), Pattern.CASE_INSENSITIVE);
        } else {
            return Pattern.compile(newRegex.toString());
        }
    }
}
