/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.common.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.UnsupportedCharsetException;
import java.util.Arrays;

/**
 * Tasks of Class StringInputStream.
 * <ol>
 * <li>This is a Low Memory Implementation to convert a String in a InputStream</li>
 * <li>It converts only CHAR_BUFFER_SIZE chars at once</li>
 * </ol>
 * 
 * @author reiprecht
 */
public class StringInputStream extends InputStream {

    private static final int CHAR_BUFFER_SIZE = 16;
    private String m_content;
    private int m_index = -1;
    private CharsetEncoder m_charsetEncoder;
    private CharBuffer m_charBuffer;
    private ByteBuffer m_byteBuffer;
    private boolean m_closed = false;

    public StringInputStream(String pContent, String pCharset) throws UnsupportedEncodingException {
        m_content = pContent;
        //Not available in JDK 1.4
        //if (pCharset == null) pCharset = Charset.defaultCharset().name();
        if (pCharset == null) {
            pCharset = System.getProperty("file.encoding");
        }
        try {
            m_charsetEncoder = Charset.forName(pCharset).newEncoder();
        } catch (UnsupportedCharsetException e) {
            throw new UnsupportedEncodingException(e.getCharsetName());
        }
        m_charsetEncoder.onMalformedInput(CodingErrorAction.REPLACE)
                .onUnmappableCharacter(CodingErrorAction.REPLACE);
        // Buffer Allocation
        m_charBuffer = CharBuffer.allocate(CHAR_BUFFER_SIZE);
        m_byteBuffer = ByteBuffer.allocate(((int) m_charsetEncoder.maxBytesPerChar()) * CHAR_BUFFER_SIZE);
    }

    public int read() throws IOException {
        try {
            try {
                byte b = m_byteBuffer.get();
                if (b != 0) {
                    return b;
                }
            } catch (Exception e) {
                //TODO: ignored?
            }
            m_byteBuffer.position(0);
            Arrays.fill(m_byteBuffer.array(), (byte) 0);
            m_charBuffer.position(0);
            fillCharBuffer();
            m_charBuffer.position(0);
            m_charsetEncoder.encode(m_charBuffer, m_byteBuffer, false);
            m_byteBuffer.position(0);

            byte b = m_byteBuffer.get();
            if (b != 0) {
                return b;
            }
            return -1;
        } catch (IndexOutOfBoundsException e) {
            return -1;
        }
    }

    private void fillCharBuffer() {
        if (m_closed) {
            throw new IndexOutOfBoundsException();
        }
        int j = -1;
        try {
            while ((++j) < CHAR_BUFFER_SIZE) {
                m_charBuffer.put(m_content.charAt(++m_index));
            }
        } catch (IndexOutOfBoundsException e) {
            m_charBuffer.limit(j);
            m_closed = true;
        }
    }

    public int available() throws IOException {
        return m_content.length();
    }

    public synchronized void reset() throws IOException {
        m_closed = false;
        m_index = -1;
        m_byteBuffer.clear();
        m_charBuffer.clear();
        Arrays.fill(m_byteBuffer.array(), (byte) 0);
    }
}
