/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 24.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.io.Writer;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.pool.KeyedPoolableObjectFactory;
import org.apache.commons.pool.impl.GenericKeyedObjectPool;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.NodeFilter;

/**
 * Class DOMSerializer is a small utility calls responsible for serializing xml documents to output streams.
 * <p/>
 * Usage:
 * <p/>
 * Use DOMSerializer.createNew() to create a new instance of a DOMSerializier and then use one of the serialize methods.
 * <p/>
 * A DOMSerializer shall be never hold as member variable, please create them only local.
 *
 * @author Marko Reiprecht
 *         Created on 24.03.2008
 * @version $Revision: $
 */
public abstract class DOMSerializer {

    // ------------------------------ FIELDS ------------------------------

    private static SimpleCopyTransformerPool copyTransformerPool = new SimpleCopyTransformerPool();

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method creates a new DOMSerializer.
     *
     * @return DOMSerializer.
     */
    public static DOMSerializer createNew() {
        return createNew(null);
    }

    /**
     * Method creates a new DOMSerializer.
     *
     * @param options the options.
     * @return DOMSerializer.
     */
    public static DOMSerializer createNew(final DOMSerializerOptions options) {
        return new PoolingDOMSerializer(options == null ? DOMSerializerOptions.getDefaultOptions() : options,
                                        copyTransformerPool);
    }

    /**
     * Method creates a new copy Transformer.
     *
     * @param options the options.
     * @return Transformer.
     * @throws DOMSerializerException if the copy transformer can't be created.
     */
    protected static Transformer createTransformer(final DOMSerializerOptions options) throws DOMSerializerException {
        try {
            TransformerFactory fac = TransformerFactory.newInstance();
            return applyTransformerOptions(fac.newTransformer(), options);
        } catch (TransformerConfigurationException e) {
            throw new DOMSerializerException(e);
        }
    }

    /**
     * Method applys the DOMSericalizerOptions to the copy tansformer.
     *
     * @param transformer the transformer.
     * @param options     the options.
     * @return Transformer.
     */
    private static Transformer applyTransformerOptions(final Transformer transformer,
                                                       final DOMSerializerOptions options) {
        transformer.setOutputProperty(OutputKeys.METHOD, "xml");
        transformer.setOutputProperty(OutputKeys.ENCODING, options.getEncoding());
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
                                      options.isOmitXMLDeclaration() ? "yes" : "no");
        transformer.setOutputProperty(OutputKeys.INDENT, options.isIndent() ? "yes" : "no");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount",
                                      String.valueOf(options.getIndentAmount()));
        return transformer;
    }

    /**
     * Method removes the indent text nodes from the dom tree.
     * <p/>
     * The semantic of this "cleanup" method is to remove any text nodes, which has element silblings. So if the xml
     * schemata contains nodes, which support "mixed-content" this method removes the mixed-content text nodes. Please
     * use this method only in Situations where you know that the algorithm works.
     * 
     * @param document
     *            the document.
     * @return Document.
     */
    public static Document removeIndentTextNodes(final Document document) {
        new DocumentTraverser().traverseDepthFirst(document.getDocumentElement(), new DocumentVisitorAdapter() {
            public void visit(final Node node) {
                check(node.getPreviousSibling(), node);
                check(node.getNextSibling(), node);
            }

            private void check(final Node silb, final Node node) {
                if (silb != null
                    && Node.ELEMENT_NODE == silb.getNodeType()) {
                    node.getParentNode().removeChild(node);
                }
            }

            public short acceptNode(final Node n) {
                switch (n.getNodeType()) {
                    case Node.TEXT_NODE:
                        return NodeFilter.FILTER_ACCEPT;
                    default:
                        return NodeFilter.FILTER_SKIP;
                }
            }
        });
        return document;
    }

    /**
     * Method serializes the xml document to the writer.
     *
     * @param document the document.
     * @param out      the out.
     * @throws DOMSerializerException if error occurs.
     */
    public void serialize(final Document document, final Writer out) throws DOMSerializerException {
        serialize((Node) document, out);
    }

    /**
     * Method serializes the document to the output stream.
     *
     * @param document the document.
     * @param out      the out.
     * @throws DOMSerializerException if error occurs.
     */
    public void serialize(final Document document, final OutputStream out) throws DOMSerializerException {
        serialize((Node) document, out);
    }

    /**
     * Method serializeToBytes converts the document to a byte[].
     *
     * @param document the document.
     * @return byte[].
     * @throws DOMSerializerException if error occurs.
     */
    public byte[] serializeToBytes(final Document document) throws DOMSerializerException {
        return serializeToBytes((Node) document);
    }

    /**
     * Method serializeToBytes converts the node to a byte[].
     *
     * @param node the node.
     * @return byte[].
     * @throws DOMSerializerException if error occurs.
     */
    public byte[] serializeToBytes(final Node node) throws DOMSerializerException {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        serialize(node, bout);
        return bout.toByteArray();
    }

    /**
     * Method serializes the node to the output stream.
     *
     * @param node the dom node.
     * @param out  the out.
     * @throws DOMSerializerException if error occurs.
     */
    public abstract void serialize(Node node, OutputStream out) throws DOMSerializerException;

    /**
     * Method serializeToString converts the document to a string.
     *
     * @param document the document.
     * @return String.
     * @throws DOMSerializerException if error occurs.
     */
    public String serializeToString(final Document document) throws DOMSerializerException {
        return serializeToString((Node) document);
    }

    /**
     * Method serializeToString converts the node to a string.
     *
     * @param node the node.
     * @return String.
     * @throws DOMSerializerException if error occurs.
     */
    public String serializeToString(final Node node) throws DOMSerializerException {
        StringWriter writer = new StringWriter();
        serialize(node, writer);
        return writer.toString();
    }

    /**
     * Method serializes the xml node to the writer.
     *
     * @param node the dom node.
     * @param out  the out.
     * @throws DOMSerializerException if error occurs.
     */
    public abstract void serialize(Node node, Writer out) throws DOMSerializerException;

    // -------------------------- INNER CLASSES --------------------------

    /**
     * Class PoolingDOMSerializer uses a small transfomer pool to get a transformer for the xml document.
     *
     * @author Marko Reiprecht
     *         Created on 23.04.2008
     * @version $Revision: $
     */
    private static class PoolingDOMSerializer extends DOMSerializer {

        private DOMSerializerOptions m_options;
        private SimpleCopyTransformerPool m_pool;

        /**
         * Constructor creates a new instance.
         *
         * @param domSerializerOptions the domSerializerOptions.
         * @param pool                 the pool.
         */
        public PoolingDOMSerializer(final DOMSerializerOptions domSerializerOptions,
                                    final SimpleCopyTransformerPool pool) {
            m_options = domSerializerOptions;
            m_pool = pool;
        }

        /**
         * @see DOMSerializer#serialize(Node, Writer)
         */
        public void serialize(final Node node, final Writer out) throws DOMSerializerException {
            Transformer transformer = m_pool.acquire(m_options);
            try {
                transformer.transform(new DOMSource(checkNamespace(node)), new StreamResult(out));
            } catch (TransformerException e) {
                throw new DOMSerializerException(e);
            } finally {
                m_pool.release(transformer, m_options);
            }
        }

        /**
         * @see DOMSerializer#serialize(Node, OutputStream)
         */
        public void serialize(final Node node, final OutputStream out) throws DOMSerializerException {
            Transformer transformer = m_pool.acquire(m_options);
            try {
                transformer.transform(new DOMSource(checkNamespace(node)), new StreamResult(out));
            } catch (TransformerException e) {
                throw new DOMSerializerException(e);
            } finally {
                m_pool.release(transformer, m_options);
            }
        }

        /**
         * Method check namespaces.
         *
         * @param node the node.
         * @return Node.
         */
        private Node checkNamespace(final Node node) {
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    if (node != node.getOwnerDocument().getDocumentElement()) {
                        // following is only needed in jdk 1.4
                        // if node to serialize is an element we have to check the namespace declarations
                        // todo: and to resolve all namespaces used :-( maybe better solutions here ?
                        // todo: open to implement:
                        /*
                        new DocumentTraverser().traverseDepthFirst(node, new DocumentVisitorAdapter() {

                            public void preVisit(final Element element) {

                            }

                            public void postVisit(final Element element) {

                            }

                            public void visit(final Attr attr) {

                            }
                        });*/
                    }
                    break;
            }
            return node;
        }
    }

    /**
     * Class SimpleCopyTransformerPool.
     *
     * @author Marko Reiprecht
     *         Created on 30.05.2008
     * @version $Revision: $
     */
    private static class SimpleCopyTransformerPool {

        /**
         * The Logger.
         */
        private static final Log LOG = LogFactory.getLog(DOMSerializer.SimpleCopyTransformerPool.class);
        private GenericKeyedObjectPool m_transformerPool;

        /**
         * Constructor creates a new instance.
         */
        private SimpleCopyTransformerPool() {
            GenericKeyedObjectPool.Config config = new GenericKeyedObjectPool.Config();
            // max number of total created documentbuilders for all possible options.
            config.maxTotal = 20;
            // max number of idle builders per option
            config.maxIdle = 5;
            // max number of active builder per option
            config.maxActive = 20;
            // don't grow, fail, if the pool is exhausted.
            // Blocking can lead to bad error states, so this poll will fail,
            // if more than 20 parsers are parallel in use.
            config.whenExhaustedAction = GenericKeyedObjectPool.WHEN_EXHAUSTED_GROW;
            // but only wait for maximal one second befor throw an exception
            // wait is obsolete if exhausted fail is configured.
            //config.maxWait = 1000;
            m_transformerPool = new GenericKeyedObjectPool(new TransformerKeyedPoolableObjectFactory(), config);
        }

        /**
         * Method acquires a new Transformer from the pool.
         *
         * @param options the options.
         * @return Transformer.
         */
        public Transformer acquire(final DOMSerializerOptions options) {
            try {
                return (Transformer) m_transformerPool.borrowObject(options);
            } catch (Exception e) {
                throw new DOMSerializerException("CopyTransformer can not retrieved from pool", e);
            }
        }

        /**
         * Method releases a transformer back in the pool.
         *
         * @param transformer the transformer.
         * @param options     the options.
         */
        public void release(final Transformer transformer, final DOMSerializerOptions options) {
            try {
                m_transformerPool.returnObject(options, transformer);
            } catch (Exception e) {
                if (LOG.isWarnEnabled()) { LOG.warn("Transformer could not returned in pool"); }
            }
        }
    }

    /**
     * Class TransformerKeyedPoolableObjectFactory.
     *
     * @author Marko Reiprecht
     *         Created on 30.05.2008
     * @version $Revision: $
     */
    private static class TransformerKeyedPoolableObjectFactory implements KeyedPoolableObjectFactory {

        /**
         * Method makes a transformer.
         *
         * @param options the options.
         * @return Object.
         * @throws Exception if error occurs.
         */
        public Object makeObject(final Object options) throws Exception {
            return DOMSerializer.createTransformer((DOMSerializerOptions) options);
        }

        /**
         * Method destroys the transformer.
         *
         * @param options     the options.
         * @param transformer the transformer.
         * @throws Exception if error occurs.
         */
        public void destroyObject(final Object options, final Object transformer) throws Exception {
            ((Transformer) transformer).setErrorListener(null);
            ((Transformer) transformer).setURIResolver(null);
        }

        /**
         * Method validates the transformer.
         *
         * @param options     the options.
         * @param transformer the transformer.
         * @return boolean.
         */
        public boolean validateObject(final Object options, final Object transformer) {
            return true;
        }

        /**
         * Method activates the transformer.
         *
         * @param options     the options.
         * @param transformer the transformer.
         * @throws Exception if error occurs.
         */
        public void activateObject(final Object options, final Object transformer) throws Exception {
            // nothing
        }

        /**
         * Method passivates the transformer.
         *
         * @param options     the options.
         * @param transformer the transformer.
         * @throws Exception if error occurs.
         */
        public void passivateObject(final Object options, final Object transformer) throws Exception {
            ((Transformer) transformer).setErrorListener(null);
            ((Transformer) transformer).setURIResolver(null);
        }
    }
}
