/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 05.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;

/**
 * Class SimpleNamespaceContext is a simple NamespaceContext implementation.
 * <p/>
 * It is possible to register more as one prefix for a namespace. Internally it uses to maps and a prefix set to
 * fullfill the interface specification.
 * <p/>
 * If your xml contains a default namespace than you have to register it with the empty string or
 * {@link javax.xml.XMLConstants.DEFAULT_NS_PREFIX}.
 * <p/>
 * It is not necessary to register the xml or xmlns namespace prefixes.
 * <p/>
 * Usage:<br/>
 * <code>
 * new SimpleNamespaceContext()
 * .add("myns","http://tempuri.org/myns")
 * .add("my2ns","http://tempuri.org/my2ns");
 * </code>
 * 
 * @author Marko Reiprecht Created on 05.03.2008
 * @version $Revision: 1.1.2.2 $
 */
public class SimpleNamespaceContext implements NamespaceContext {

    // ------------------------------ FIELDS ------------------------------

    private Map m_namespacePrefixMap = new HashMap();
    private Map m_prefixNamespaceMap = new HashMap();

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance.
     */
    public SimpleNamespaceContext() {
    }

    // ------------------------ INTERFACE METHODS ------------------------

    // --------------------- Interface NamespaceContext ---------------------

    /**
     * Method gets a namespace uri for a prefix.
     * 
     * @see javax.xml.namespace.NamespaceContext#getNamespaceURI(String)
     */
    public String getNamespaceURI(final String prefix) {
        if (prefix == null) {
            throw new IllegalArgumentException("prefix must not null");
        }
        String nsuri = (String) m_prefixNamespaceMap.get(prefix);
        if (nsuri == null) {
            // following is required to fullfill the interface specification
            if (XMLConstants.DEFAULT_NS_PREFIX.equals(prefix)) {
                nsuri = XMLConstants.NULL_NS_URI;
            } else if (XMLConstants.XML_NS_PREFIX.equals(prefix)) {
                nsuri = XMLConstants.XML_NS_URI;
            } else if (XMLConstants.XMLNS_ATTRIBUTE.equals(prefix)) {
                nsuri = XMLConstants.XMLNS_ATTRIBUTE_NS_URI;
            } else {
                nsuri = XMLConstants.NULL_NS_URI;
            }
        }
        return nsuri;
    }

    /**
     * Method gets the Prefix for a namespace uri.
     * 
     * @see javax.xml.namespace.NamespaceContext#getPrefix(String)
     */
    public String getPrefix(final String namespaceURI) {
        Iterator prefixes = getPrefixes(namespaceURI);
        if (prefixes.hasNext()) {
            return (String) prefixes.next();
        }
        return null;
    }

    /**
     * Method getPrefixes for a namespace uri.
     * 
     * @see javax.xml.namespace.NamespaceContext#getPrefixes(String)
     */
    public Iterator getPrefixes(final String namespaceURI) {
        if (namespaceURI == null) {
            throw new IllegalArgumentException("namespaceURI must be not null");
        }
        Set prefixes = (Set) m_namespacePrefixMap.get(namespaceURI);
        if (prefixes == null) {
            // following is required to fullfill the interface specification
            if (XMLConstants.XML_NS_URI.equals(namespaceURI)) {
                return Collections.singleton(XMLConstants.XML_NS_PREFIX).iterator();
            } else if (XMLConstants.XMLNS_ATTRIBUTE_NS_URI.equals(namespaceURI)) {
                return Collections.singleton(XMLConstants.XMLNS_ATTRIBUTE).iterator();
            } else {
                return Collections.EMPTY_SET.iterator();
            }
        }
        return prefixes.iterator();
    }

    /**
     * @return all prefixes within the context.
     */
    public Iterator getPrefixes() {
        return m_prefixNamespaceMap.keySet().iterator();
    }

    /**
     * @return all namespace uris within the context.
     */
    public Iterator getNamespaceURIs() {
        return m_namespacePrefixMap.keySet().iterator();
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method adds a prefix to namespace mapping.
     * 
     * @param prefix
     *            the prefix.
     * @param namespace
     *            the namespace.
     * @return SimpleNamespaceContext.
     */
    public SimpleNamespaceContext add(final String prefix, final String namespace) {
        m_prefixNamespaceMap.put(prefix, namespace);
        if (!m_namespacePrefixMap.containsKey(namespace)) {
            m_namespacePrefixMap.put(namespace, new HashSet());
        }
        ((Set) m_namespacePrefixMap.get(namespace)).add(prefix);
        return this;
    }

    /**
     * Adds all prefixes and namespaces from the given context to this context.
     * 
     * @param ctx
     *            another simple namespace context.
     * @return this context.
     */
    public SimpleNamespaceContext addAll(final SimpleNamespaceContext ctx) {
        Iterator it = ctx.getPrefixes();
        while (it.hasNext()) {
            String prefix = (String) it.next();
            add(prefix, ctx.getNamespaceURI(prefix));
        }
        return this;
    }

    /**
     * Adds all prefixes and namespaces from the given Map<prefix,namespace> to this context.
     * 
     * @param prefixNamespaceMap
     *            a map with prefix to namespace mappings.
     * @return this context.
     */
    public SimpleNamespaceContext addAll(final Map prefixNamespaceMap) {
        Iterator it = prefixNamespaceMap.keySet().iterator();
        while (it.hasNext()) {
            String prefix = (String) it.next();
            add(prefix, (String) prefixNamespaceMap.get(prefix));
        }
        return this;
    }

}
