/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.04.2008
 ******************************************************************************/
package org.n52.security.common.xml;

import java.io.IOException;
import java.io.Writer;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.Namespace;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import org.xml.sax.InputSource;

/**
 * Class XMLStreamChunkReader reads an xml document with an internal xml stream reader
 * and as chunks and calls a subsets of XMLStreamChunkHandler instances.
 * <p/>
 * Xml elements for which no handler is defined are simply skipped.
 *
 * @author Marko Reiprecht
 *         Created on 18.04.2008
 * @version $Revision: $
 */
public class XMLChunkReader {

    /**
     * Contains a map from qnames to handlers.
     */
    private Map m_partialStreamHandler = new HashMap();

    /**
     * Method registers a PartialStreamHandler for a Qname.
     * 
     * @param qname
     *            the qname.
     * @param handler
     *            the handler.
     * @return this.
     */
    public XMLChunkReader addHandler(final QName qname, final XMLChunkHandler handler) {
        m_partialStreamHandler.put(qname, handler);
        return this;
    }

    /**
     * Method gets the Handler for the qname.
     * 
     * @param qname
     *            the qname.
     * @return XmlPartialStreamHandler.
     */
    public XMLChunkHandler getHandler(final QName qname) {
        return (XMLChunkHandler) m_partialStreamHandler.get(qname);
    }

    /**
     * Method reads an xml source.
     * 
     * @param source
     *            the source.
     * @throws XMLStreamException
     *             if error occurs.
     */
    public void read(final InputSource source) throws XMLStreamException {
        // create event reader from the input source.
        XMLEventReader reader = createEventReader(source);
        // create event factory to create new xml events
        XMLEventFactory fac = XMLEventFactory.newInstance();
        // create a dynamicaly grown writer pool
        LinkedList writerPool = new LinkedList();
        // create handler stack
        LinkedList handlerStack = new LinkedList();
        // create namespace stack
        LinkedList namespaceStack = new LinkedList();
        // create xml path stack
        LinkedList qnamePath = new LinkedList();
        try {
            while (reader.hasNext()) {
                XMLEvent evt = reader.nextEvent();
                // If this event starts the element of interest,
                // change the flag to indicate such.
                if (evt.isStartElement()) {
                    evt = handleStartElement(fac, writerPool, handlerStack, namespaceStack, qnamePath, evt);
                }
                // If we are in the element, write the events out.
                if (!handlerStack.isEmpty()) {
                    ((HandlerStackEntry)handlerStack.getFirst()).getEventWriter().add(evt);
                }
                // If this is the end of the element
                // and an handler is registered for this element, so call it
                if (evt.isEndElement()) {
                    handleEndElement(handlerStack, namespaceStack, qnamePath, evt);
                }
            }
        } finally {
            reader.close();
        }
    }

    /**
     * Handles the start element event.
     * It checks for registered handlers and fires the appropriate events on them.
     * 
     * @param fac
     *            a event factory.
     * @param writerPool
     *            the writer pool.
     * @param handlerStack
     *            the handler stack.
     * @param namespaceStack
     *            the namespace stack.
     * @param qnamePath
     *            the current xpath location.
     * @param evt
     *            the current read event.
     * @return a new (repaired) event.
     * @throws XMLStreamException
     */
    private XMLEvent handleStartElement(final XMLEventFactory fac,
                                        final LinkedList writerPool,
                                        final LinkedList handlerStack,
                                        final LinkedList namespaceStack,
                                        final LinkedList qnamePath,
                                        final XMLEvent evt) throws XMLStreamException {
        StartElement elem = (StartElement) evt;
        NamespaceStackEntry namespaceEntry = new NamespaceStackEntry(elem.getNamespaces(), namespaceStack,
                fac);
        QName qName = elem.getName();
        qnamePath.addLast(qName);
        XMLChunkHandler handler = getHandler(qName);
        if (handler != null) {
            List currentPath = Collections.unmodifiableList(qnamePath);
            if (!handlerStack.isEmpty()) {
                HandlerStackEntry entry = (HandlerStackEntry) handlerStack.getFirst();
                entry.getHandler().handleChunkBeforChildHandler(
                        currentPath,
                        entry.getQname(),
                        entry.getWriterContent(),
                        qName,
                        handler);
            }
            HandlerStackEntry newActive = new HandlerStackEntry(handler, qName, writerPool);
            handlerStack.addFirst(newActive);
            newActive.getHandler().handleChunkStart(currentPath, qName);
        }
        // we have to append the start element with all available namespaces
        // otherwise the element is not correct
        // and XMLOutputFactory.IS_REPAIRING_NAMESPACES, does not work,
        // because some elements are skiped
        // it also optimizes the namespace declarations, no namespace declarations are repeated!
        return fac.createStartElement(elem.getName(), elem.getAttributes(), handler != null ? namespaceEntry
                .getNamespaces() : namespaceEntry.getLocalNamespaces());
    }

    /**
     * Handler function of the end element event.
     * It calls the appropriate events on the registered handlers.
     * 
     * @param handlerStack
     *            the handler stack.
     * @param namespaceStack
     *            the namespace stack.
     * @param qnamePath
     * @param evt
     *            the current end element event.
     * @throws XMLStreamException
     */
    private void handleEndElement(final LinkedList handlerStack,
                                  final LinkedList namespaceStack,
                                  final LinkedList qnamePath,
                                  final XMLEvent evt) throws XMLStreamException {
        namespaceStack.removeFirst();
        if (!handlerStack.isEmpty()) {
            List currentPath = Collections.unmodifiableList(qnamePath);
            EndElement elem = (EndElement) evt;
            HandlerStackEntry handlerentry = (HandlerStackEntry) handlerStack.getFirst();
            if (elem.getName().equals(handlerentry.getQname())) {
                handlerStack.removeFirst();
                handlerentry.getHandler().handleChunkEnd(currentPath, handlerentry.getQname(),
                        handlerentry.getWriterContent());

                if (!handlerStack.isEmpty()) {
                    HandlerStackEntry parentEntry = (HandlerStackEntry) handlerStack.getFirst();
                    parentEntry.getHandler().handleChunkAfterChildHandler(currentPath, parentEntry.getQname(),
                            parentEntry.getWriterContent(), handlerentry.getQname(),
                            handlerentry.getHandler());
                }
                // release the used writer
                handlerentry.release();
            }
        }
        qnamePath.removeLast();
    }

    /**
     * Creates the XMLEventReader working on the given inputsource.
     * 
     * @param source
     *            the inputsource.
     * @return a XMLEventReader connected to the inputsource.
     * @throws XMLStreamException
     */
    private XMLEventReader createEventReader(final InputSource source)
            throws XMLStreamException {
        // Create the input reader to read the source.
        XMLInputFactory inFactory = XMLInputFactory.newInstance();
        inFactory.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, Boolean.TRUE);
        // in jdk 1.6 this method is throws an Malformed URL exception, so we must use
        // the plain methods
        // XMLEventReader reader = inFactory.createXMLEventReader(new SAXSource(source));
        final XMLEventReader reader;
        if (source.getByteStream() != null) {
            if (source.getEncoding() != null && source.getEncoding().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getByteStream(), source.getEncoding());
            } else if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getByteStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getByteStream());
            }
        } else if (source.getCharacterStream() != null) {
            if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getCharacterStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getCharacterStream());
            }
        } else {
            throw new IllegalArgumentException("InputSource holds no valid input");
        }
        return reader;
    }

    /**
     * Class NamespaceStackEntry encapsulates the namespaces available for each element. The namespaces must be managed
     * in stack and appended to the startelement of each chunk, so that the xml is legal.
     * 
     * @author Marko Reiprecht Created on 19.04.2008
     * @version $Revision: $
     */
    private static final class NamespaceStackEntry {
        /**
         * Stores prefixes to namespace names.
         */
        private Map m_localNamespaces = new HashMap();
        private Map m_globalNamespaces;

        /**
         * Constructor creates a new instance.
         * 
         * @param elementNamespaces
         *            the elementNamespaces.
         * @param namespaceStack
         *            the namespaceStack.
         * @param fac
         */
        private NamespaceStackEntry(final Iterator elementNamespaces,
                final LinkedList namespaceStack, final XMLEventFactory fac) {
            parseNamespaces(elementNamespaces, namespaceStack, fac);
            namespaceStack.addFirst(this);
        }

        /**
         * Responsible for .
         * 
         * @param elementNamespaces
         * @param namespaceStack
         * @param fac
         */
        private void parseNamespaces(final Iterator elementNamespaces,
                                     final LinkedList namespaceStack,
                                     final XMLEventFactory fac) {
            addToLocalNamespaces(elementNamespaces);
            if (!namespaceStack.isEmpty()) {
                // don't copy only link, because it are the same namespaces
                m_globalNamespaces = ((NamespaceStackEntry)namespaceStack.getFirst()).m_globalNamespaces;
            }
            if (!m_localNamespaces.isEmpty() && m_globalNamespaces == null) {
                // if it is the first element!
                m_globalNamespaces = m_localNamespaces;
                // this fixes a bug in the codehaus stax implementation, where xml:lang attributes not correctly writen
                // if the xmlns:xml=".." is not emited!
                if (!m_globalNamespaces.containsKey(XMLConstants.XML_NS_PREFIX)) {
                    m_globalNamespaces.put(XMLConstants.XML_NS_PREFIX, fac.createNamespace(XMLConstants.XML_NS_PREFIX,
                            XMLConstants.XML_NS_URI));
                }
            } else if (!m_localNamespaces.isEmpty()) {
                // remove duplicated local namespaces
                for (Iterator iterator = m_localNamespaces.keySet().iterator(); iterator.hasNext();) {
                    String prefix = (String) iterator.next();
                    // test for duplicated prefix + uri
                    // be carefull if you check namespace.equals(namespace) and the provider (till jdk6_12) has wrong
                    // implemented the equals method than the test fails
                    // so here is excplicite checked for prefix and namespace
                    if (m_globalNamespaces.containsKey(prefix)
                            && ((Namespace)m_globalNamespaces.get(prefix)).getNamespaceURI().equals(
                                    ((Namespace)m_localNamespaces.get(prefix)).getNamespaceURI())) {
                        iterator.remove();
                    }
                }
                if (!m_localNamespaces.isEmpty()) {
                    m_globalNamespaces = new HashMap(m_globalNamespaces);
                    m_globalNamespaces.putAll(m_localNamespaces);
                }
            }
        }

        /**
         * adds a all namespaces to the local namespaces.
         * 
         * @param namespaces
         *            iterator of namespaces.
         */
        private void addToLocalNamespaces(final Iterator namespaces) {
            while (namespaces.hasNext()) {
                Namespace namespace = (Namespace) namespaces.next();
                m_localNamespaces.put(namespace.getPrefix(), namespace);
            }
        }

        /**
         * Gets on the element declared local namespaces.
         * 
         * @return
         */
        public Iterator getLocalNamespaces() {
            return m_localNamespaces.values().iterator();
        }

        /**
         * Method gets the namespaces.
         * 
         * @return Iterator.
         */
        Iterator getNamespaces() {
            return m_globalNamespaces == null ? Collections.EMPTY_LIST.iterator() : m_globalNamespaces.values().iterator();
        }
    }

    /**
     * Class WriterPoolEntry encapsulates eventwriters, which are used to generate the xmlchunks for each handler. They
     * can be reused for different handlers, therefor its structured as a pool entry.
     * 
     * @author Marko Reiprecht Created on 19.04.2008
     * @version $Revision: $
     */
    private static final class WriterPoolEntry {

        private ResettableStringWriter m_stringWriter;
        private XMLEventWriter m_eventWriter;
        private LinkedList m_poolQueue;
        private XMLOutputFactory m_eventWriterFactory;

        /**
         * Constructor creates a new instance.
         * 
         * @param poolQueue
         *            the poolQueue.
         * @throws XMLStreamException
         *             if error occurs.
         */
        private WriterPoolEntry(final LinkedList poolQueue) throws XMLStreamException {
            // Create an output writer. The writer uses the resettable
            // buffer so we can use the same writer continuously.
            XMLOutputFactory outFactory = XMLOutputFactory.newInstance();
            // the flag is needed so that namespaces are appended.
            // the flag works only for element namespaces, not for attribute namespaces,
            // therefor the namespaces are managed in an extra stack.
            // outFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
            m_stringWriter = new ResettableStringWriter();
            m_eventWriterFactory = outFactory;
            m_poolQueue = poolQueue;
        }

        /**
         * Method gets the stringWriter.
         * 
         * @return ResettableStringWriter.
         */
        public ResettableStringWriter getStringWriter() {
            return m_stringWriter;
        }

        /**
         * Method gets the eventWriter.
         * 
         * @return XMLEventWriter.
         */
        public XMLEventWriter getEventWriter() throws XMLStreamException {
            if (m_eventWriter == null) {
                m_stringWriter.reset();
                m_eventWriter = m_eventWriterFactory.createXMLEventWriter(m_stringWriter);
            }
            return m_eventWriter;
        }

        /**
         * Method release the pool entry and make it available for other handlers.
         */
        public void release() throws XMLStreamException {
            m_eventWriter.close();
            m_eventWriter = null;
            m_stringWriter.reset();
            m_poolQueue.addLast(this);
        }
    }

    /**
     * Class HandlerStackEntry.
     * 
     * @author Marko Reiprecht Created on 18.04.2008
     * @version $Revision: $
     */
    private static final class HandlerStackEntry {

        private XMLChunkHandler m_handler;
        private QName m_qname;
        private WriterPoolEntry m_poolEntry;

        /**
         * Constructor creates a new instance.
         * 
         * @param handler
         *            the handler.
         * @param qname
         *            the qname.
         * @param writerPool
         *            the pooled event writers.
         * @throws javax.xml.stream.XMLStreamException
         *             if a new writer pool instance can't be created.
         */
        private HandlerStackEntry(final XMLChunkHandler handler, final QName qname,
                final LinkedList writerPool) throws XMLStreamException {
            m_handler = handler;
            m_qname = qname;
            if (writerPool.isEmpty()) {
                m_poolEntry = new WriterPoolEntry(writerPool);
            } else {
                m_poolEntry = (WriterPoolEntry) writerPool.removeFirst();
            }
        }

        /**
         * Method gets the handler.
         * 
         * @return XmlPartialStreamHandler.
         */
        public XMLChunkHandler getHandler() {
            return m_handler;
        }

        /**
         * Method gets the qname.
         * 
         * @return QName.
         */
        public QName getQname() {
            return m_qname;
        }

        /**
         * Method gets the eventWriter.
         * 
         * @return XMLEventWriter.
         */
        public XMLEventWriter getEventWriter() throws XMLStreamException {
            return m_poolEntry.getEventWriter();
        }

        /**
         * Method release.
         */
        public void release() throws XMLStreamException {
            m_poolEntry.release();
        }

        /**
         * Method gets the currently written content in the buffer.
         * 
         * @return String.
         * @throws XMLStreamException
         *             if error occurs.
         */
        public String getWriterContent() throws XMLStreamException {
            m_poolEntry.getEventWriter().flush();
            return m_poolEntry.getStringWriter().getString();
        }
    }

    /**
     * A Writer that buffers the data in memory as a string. The buffer can be reset at any time and the contents of the
     * buffer will be returned.
     */
    private static final class ResettableStringWriter extends Writer {

        /**
         * The internal buffer used to store the data written to this stream.
         */
        private StringBuilder m_buffer = new StringBuilder(1024);

        /**
         * The maximum number of chars that this writer will reuse. (2MB)
         */
        private int m_maxChars = 1024 * 1024;

        /**
         * Constructor creates a new instance.
         */
        private ResettableStringWriter() {
        }

        /**
         * Constructor creates a new instance.
         * 
         * @param maxChars
         *            the maxChars this buffer will reuse.
         */
        private ResettableStringWriter(final int maxChars) {
            m_maxChars = maxChars;
        }

        public void close() throws IOException {
            // no op
        }

        public void flush() throws IOException {
            // no op
        }

        public void write(final char[] cbuf, final int off, final int len) throws IOException {
            m_buffer.append(cbuf, off, len);
        }

        /**
         * Method gets the buffer.
         * 
         * @return StringBuffer.
         */
        public String getString() {
            return m_buffer.substring(0);
        }

        /**
         * Resets the stream to a length of zero. Any data written to the stream since the last call to reset will be
         * returned.
         * 
         * @return the contents of the stream buffer
         */
        public ResettableStringWriter reset() {
            if (m_buffer.length() > m_maxChars) {
                m_buffer = new StringBuilder(1024);
            }
            m_buffer.setLength(0);
            return this;
        }
    } 
}
