/* Copyright (C) 2007 con terra GmbH (http://www.conterra.de)
 * All rights reserved
 *
 * $Id: $
 *
 * Originally created
 * by :     Marko Reiprecht
 * at :     16.11.2009 18:10:16
 */
package org.n52.security.common.xml;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.XMLConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * This utility removes all xmlns attribute declarations from the dom, but preserves indirect referenced namespaces e.g.
 * linked via attribute/element content( xsi:type="prefix:...").
 * 
 * @author Marko Reiprecht
 */
public class XMLNamespaceAttributeDeclarationRemover {
    private static final Log LOG = LogFactory.getLog(XMLNamespaceAttributeDeclarationRemover.class);
    private static final Pattern NSPREFIX_PATTERN = Pattern.compile("^([a-zA-Z0-1_\\-\\.]+):[a-zA-Z0-1_\\-\\.]+$");

    /**
     * Method removes the namespace declarations.
     * 
     * @param document
     *            the document.
     * @return Document.
     */
    public Document removeNSDecls(final Document document) {
        removeNSDecls(document.getDocumentElement());
        return document;
    }

    /**
     * Method removes all namespace attributes.
     * 
     * @param node
     *            the root.
     * @return Element.
     */
    public Element removeNSDecls(final Element node) {
        new DocumentTraverser().traverseDepthFirst(node, new DocumentVisitorAdapter() {
            private Matcher m_matcher = NSPREFIX_PATTERN.matcher("");
            /**
             * collect all found namespaces.
             */
            private Map m_foundNamespaces = new HashMap();
            /**
             * lists required namespaces (used in attribute values).
             */
            private Set m_requiredNamespacePrefixes = new HashSet();

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#visit(org.w3c.dom.Attr)
             */
            public void visit(final Attr attr) {
                String name = attr.getName();
                String value = attr.getValue();
                if (name.startsWith(XMLConstants.XMLNS_ATTRIBUTE)) {
                    m_foundNamespaces.put(name.replace(XMLConstants.XMLNS_ATTRIBUTE + ":", ""), attr);
                    attr.getOwnerElement().removeAttributeNode(attr);
                } else if (value != null && m_matcher.reset(value).matches()) {
                    m_requiredNamespacePrefixes.add(m_matcher.group(1));
                }
            }

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#postVisit(org.w3c.dom.Text)
             */
            public void visit(Text text) {
                String value = text.getNodeValue();
                if (value != null && m_matcher.reset(value).matches()) {
                    m_requiredNamespacePrefixes.add(m_matcher.group(1));
                }
            }

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#postVisit(org.w3c.dom.Element)
             */
            public void postVisit(Element element) {
                if (element == node) {
                    // end of iteration
                    if (m_requiredNamespacePrefixes.size() > 0) {
                        for (Iterator prefixes = m_requiredNamespacePrefixes.iterator(); prefixes.hasNext();) {
                            String prefix = (String) prefixes.next();
                            Attr decl = (Attr) m_foundNamespaces.get(prefix);
                            if (decl == null) {
                                // look up in parent
                                decl = findXMLNSDeclInParentNode(element, prefix);
                            }
                            if (decl != null) {
                                // append xmlns declaration to root element
                                element.setAttributeNodeNS(decl);
                            } else {
                                if (LOG.isWarnEnabled()) {
                                    LOG.warn("no namespace attribute found in sub tree for prefix <" + prefix + ">");
                                }
                            }
                        }
                    }
                }
            }

            private Attr findXMLNSDeclInParentNode(final Element element, final String prefix) {
                String attrName = XMLConstants.XMLNS_ATTRIBUTE + ":" + prefix;
                Element cur = element;
                while (cur != null) {
                    Attr attr = cur.getAttributeNode(attrName);
                    if (attr != null) {
                        // we don't remove this node, so we have to clone it
                        return (Attr) attr.cloneNode(false);
                    }
                    Node parent = cur.getParentNode();
                    if (Node.ELEMENT_NODE == parent.getNodeType()) {
                        cur = (Element) parent;
                    } else {
                        cur = null;
                    }
                }
                return null;
            }

        });
        return node;
    }
}
