/* Copyright (C) 2007 con terra GmbH (http://www.conterra.de)
 * All rights reserved
 *
 * $Id: $
 *
 * Originally created
 * by :     Marko Reiprecht
 * at :     05.02.2010 00:18:02
 */
package org.n52.security.decision;

import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicLong;
import org.apache.log4j.Logger;
import org.n52.security.authentication.LicenseReference;
import org.n52.security.authentication.SubjectUtil;

/**
 * @author Marko Reiprecht
 */
public class CachingDecisionService implements DecisionService {

    private static final Logger LOG = Logger.getLogger(CachingDecisionService.class);

    private DecisionService m_delegate;

    /**
     * <String, CacheEntry>
     */
    private final Map m_decisionCache = new HashMap();
    private final Object m_lock = new Object();
    // defaults to 5sec
    private long m_maxCacheTime = 5000;
    private int m_maxEntries = Integer.MAX_VALUE;
    private long m_lastFreeCheck = System.currentTimeMillis();
    private Timer m_cleanUpTimer;
    // used for some statistic logging
    private AtomicLong m_hitCount = new AtomicLong();
    private AtomicLong m_missCount = new AtomicLong();

    private volatile boolean m_stopped = true;

    /**
     * Creates new instance.
     * 
     * @param delegate
     */
    public CachingDecisionService() {
    }

    /**
     * Creates new instance.
     * 
     * @param delegate
     */
    public CachingDecisionService(final DecisionService delegate, final long maxCacheTimeInMsec) {
        if (delegate == null) {
            throw new IllegalArgumentException("<delegate> must not null");
        }
        m_delegate = delegate;
        m_maxCacheTime = maxCacheTimeInMsec;
    }

    public void init() {
        if (m_stopped && m_maxCacheTime > 0) {
            m_hitCount.set(0);
            m_missCount.set(0);
            m_cleanUpTimer = new Timer("decision-cache-cleanup-timer");
            m_cleanUpTimer.schedule(new TimerTask() {
                public void run() {
                    if (!m_stopped) {
                        int oldCount = 0;
                        int newCount = 0;
                        synchronized (m_lock) {
                            long now = System.currentTimeMillis();
                            boolean timeCheck = m_lastFreeCheck < now - m_maxCacheTime;
                            oldCount = m_decisionCache.size();
                            if (oldCount > 0 && (timeCheck || oldCount > m_maxEntries)) {
                                m_lastFreeCheck = now;
                                // perform clean up
                                for (Iterator entries = m_decisionCache.values().iterator(); entries.hasNext();) {
                                    CacheEntry val = (CacheEntry) entries.next();
                                    if (val == null || val.getInserted() < (now - m_maxCacheTime)) {
                                        entries.remove();
                                    }
                                }
                            }
                            newCount = m_decisionCache.size();
                        }
                        if (oldCount != newCount) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("hold " + newCount + " cache entries after cleanup (previous where "
                                        + oldCount + ") --- hit-count: <" + m_hitCount.get() + "> miss-count: <"
                                        + m_missCount.get() + ">");
                            }
                        }
                    }
                }
            }, 10000, 2000);
            m_stopped = false;
        }
    }

    public void release() {
        int lastSize = 0;
        if (!m_stopped) {
            m_stopped = true;
            m_cleanUpTimer.cancel();
            m_cleanUpTimer.purge();
            m_cleanUpTimer = null;
            synchronized (m_lock) {
                lastSize = m_decisionCache.size();
                m_decisionCache.clear();
            }
            if (LOG.isDebugEnabled()) {
                LOG.debug("closed all referenced connections");
            }
        }
        LOG.info("cache stat: lastsize: <" + lastSize + "> hit-count: <" + m_hitCount.get() + "> miss-count: <"
                + m_missCount.get() + ">");
    }

    /**
     * @return the delegate
     */
    public DecisionService getDelegate() {
        return m_delegate;
    }

    /**
     * @param delegate
     *            the delegate to set
     */
    public void setDelegate(final DecisionService delegate) {
        m_delegate = delegate;
    }

    /**
     * @return the maxCacheTime
     */
    public long getMaxCacheTime() {
        return m_maxCacheTime;
    }

    /**
     * @param maxCacheTime
     *            the maxCacheTime to set
     */
    public void setMaxCacheTime(long maxCacheTime) {
        m_maxCacheTime = maxCacheTime;
    }

    /**
     * @return the maxEntries
     */
    public int getMaxEntries() {
        return m_maxEntries;
    }

    /**
     * @param maxEntries
     *            the maxEntries to set
     */
    public void setMaxEntries(int maxEntries) {
        m_maxEntries = maxEntries;
    }

    /**
     * @see org.n52.security.decision.DecisionService#request(org.n52.security.decision.PDPRequestCollection)
     */
    public PDPResponseCollection request(final PDPRequestCollection reqCol) throws DecisionProcessingException {
        if (m_delegate == null) {
            throw new IllegalStateException("property <delegate> is not configured!");
        }
        Map allRequestsLookup = new LinkedHashMap();
        PDPRequestCollection openRequests = findNonCachedRequests(reqCol, allRequestsLookup);
        PDPResponseCollection lastResponse = null;
        if (!openRequests.isEmpty()) {
            lastResponse = getDelegate().request(openRequests);
            putInCache(lastResponse, allRequestsLookup);
        }
        // little optimization
        if (lastResponse != null && lastResponse.size() == reqCol.size()) {
            return lastResponse;
        }
        PDPResponseCollection response = new PDPResponseCollection();
        for (Iterator requests = reqCol.iterator(); requests.hasNext();) {
            PDPResponse resp = (PDPResponse) allRequestsLookup.get(requests.next());
            response.add(resp);
        }
        return response;
    }

    /**
     * Responsible for .
     * 
     * @param reqCol
     * @param allRequestsLookup
     * @return
     */
    private PDPRequestCollection findNonCachedRequests(final PDPRequestCollection reqInCol, final Map allRequestsLookup) {
        PDPRequestCollection reqcol = new PDPRequestCollection();
        for (Iterator reqIt = reqInCol.iterator(); reqIt.hasNext();) {
            PDPRequest req = (PDPRequest) reqIt.next();
            // here check for cache
            PDPResponse resp = lookUp(req);
            // if not in cache...
            if (resp == null) {
                reqcol.add(req);
            }
            // also add with null to ensure correct order!
            allRequestsLookup.put(req, resp);
        }
        return reqcol;
    }

    /**
     * Responsible for storing new found responses in the cache.
     * 
     * @param lastResponse
     * @param allRequestsLookup
     */
    private void putInCache(final PDPResponseCollection responseCol, final Map allRequestsLookup) {
        for (Iterator resIt = responseCol.iterator(); resIt.hasNext();) {
            PDPResponse resp = (PDPResponse) resIt.next();
            // put in cache...
            store(resp.getRequest(), resp);
            allRequestsLookup.put(resp.getRequest(), resp);
        }
    }

    /**
     * Responsible for .
     * 
     * @param req
     * @return
     */
    private PDPResponse lookUp(final PDPRequest req) {
        if (m_stopped || m_maxCacheTime < 0) {
            // cache disabled
            return null;
        }
        String key = getCacheKey(req);
        CacheEntry cacheEntry = null;
        synchronized (m_lock) {
            cacheEntry = (CacheEntry) m_decisionCache.get(key);
        }
        long now = System.currentTimeMillis();
        PDPResponse resp = null;
        if (cacheEntry != null
                && cacheEntry.getInserted() > (now - m_maxCacheTime)
                && (resp = cacheEntry.getResponse()) != null) {
            if (LOG.isTraceEnabled()) {
                LOG.trace("cache hit for <" + req + "> --return--> <" + resp + ">");
            }
            m_hitCount.incrementAndGet();
            // return copy to ensure that the request object is the input param!
            return new PDPResponse(resp.getDecision(), req, resp.getObligations());
        }
        m_missCount.incrementAndGet();
        return null;
    }

    /**
     * Responsible for creating a unique key from the request object.
     * 
     * @param req
     * @return
     */
    private String getCacheKey(final PDPRequest req) {
        // here we try to generate a hopefull unique "cachekey"
        // using the id of the user
        // the environment attributes,
        // and the the resource and action id
        String resource = req.getTarget().getResourceId();
        String action = req.getTarget().getActionId();
        String env = String.valueOf(req.getEnvironment().hashCode());
        String userid = SubjectUtil.getUsername(req.getTarget().getSubject());
        LicenseReference licref = SubjectUtil.getLicenseReference(req.getTarget().getSubject());
        String licId = licref != null ? licref.getLicenseId() : "";
        if (userid == null) {
            userid = "#anonymous#";
        }
        String key = new StringBuffer()
                .append("U:").append(userid)
                .append(";R:").append(resource)
                .append(";A:").append(action)
                .append(";L:").append(licId)
                .append(";E:").append(env).toString();
        if (LOG.isTraceEnabled()) {
            LOG.trace("generate hash-key <" + key + "> for request <" + req + ">");
        }
        return key;
    }

    /**
     * Responsible for .
     * 
     * @param request
     * @param resp
     */
    private void store(final PDPRequest request, final PDPResponse resp) {
        if (m_stopped || m_maxCacheTime < 0) {
            return;
        }
        // store copy to ensure no memory leeks because of pdp responses/requests!
        CacheEntry entry =
            new CacheEntry(new PDPResponse(resp.getDecision(), new PDPRequest(new Target()), resp.getObligations()));
        String key = getCacheKey(request);
        synchronized (m_lock) {
            int size = m_decisionCache.size();
            if (size > m_maxEntries) {
                // don't insert if limit is exceeded!
                return;
            }
            m_decisionCache.put(key, entry);
        }
    }

    private static class CacheEntry {
        private SoftReference m_response;
        private long m_inserted = System.currentTimeMillis();

        /**
         * Creates new instance.
         * 
         * @param response
         */
        public CacheEntry(final PDPResponse response) {
            if (response != null) {
                m_response = new SoftReference(response);
            }
        }

        /**
         * @return the inserted
         */
        public long getInserted() {
            return m_inserted;
        }

        /**
         * @return the response
         */
        public PDPResponse getResponse() {
            return (PDPResponse) (m_response != null ? m_response.get() : null);
        }
    }

}
