/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.decision;

import java.io.Serializable;
import javax.security.auth.Subject;
import org.n52.security.common.attributes.Attributes;


/**
 * The target to clearly identify an issue, for which the rights should be retrieved.
 *
 * @author Bastian Schaeffer
 * @author Marko Reiprecht
 * @version $Revision: 1748 $
 */
public class Target implements Serializable {

    // ------------------------------ FIELDS ------------------------------
    private static final String ID_ATTRIBUTE = "id";
    /**
     * the action id attribute key.
     */
    public static final String ACTION_ID = ID_ATTRIBUTE;
    /**
     * the resource id attribute key.
     */
    public static final String RESOURCE_ID = ID_ATTRIBUTE;
    /**
     * @deprecated the type attribute is not longer supported and will be removed in the next release.
     */
    public static final String TYPE_UNKNOWN = "[unknown]";
    private static final long serialVersionUID = 7751333779609996963L;


    /**
     * The resource type attribute key, this is only introduced for backwards compartibility,
     * and will be removed in the next release.
     */
    protected static final String RESOURCE_TYPE = "type";
    /**
     * Subject information of target.
     */
    protected final Subject m_subject;
    /**
     * Attributes of the resource element.
     */
    protected final Attributes m_resourceAttributes = new Attributes();
    /**
     * Attributes of the action element.
     */
    protected final Attributes m_actionAttributes = new Attributes();

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new empty target instance.
     */
    public Target() {
        this(new Subject(), null, null);
    }

    /**
     * Constructor creates a new instance with the given resource and action id, but with an empty subject.
     *
     * @param resourceId the resourceId.
     * @param actionId   the actionId.
     * @throws NullPointerException if error occurs.
     */
    public Target(final String resourceId, final String actionId)
            throws NullPointerException {
        this(new Subject(), resourceId, actionId);
    }

    /**
     * @param subject    the user who requests
     * @param resourceId the resource, e.g. a layer
     * @param actionId   the action, which should be performend. e.g. GetMap
     * @throws IllegalArgumentException if subject is null.
     */
    public Target(final Subject subject, final String resourceId, final String actionId)
            throws IllegalArgumentException {
        if (subject == null) {
            throw new IllegalArgumentException("<subject> must not null");
        }
        m_subject = subject;
        if (resourceId != null) {
            setResourceId(resourceId);
        }
        if (actionId != null) {
            setActionId(actionId);
        }
    }

    /**
     * @param subject  the subject, who ask for a resource e.g. user
     * @param resource the resource, which is asked for
     * @param action   the action, which should executed
     * @param type     a special type use {@link #TYPE_UNKNOWN} for unknown types
     * @throws IllegalArgumentException if subject is null.
     * @deprecated the type attribute is no longer directly supported,
     *             please use an application specific "type" attriubte in the ressource or action attribute collections.
     */
    public Target(final Subject subject, final String resource,
                  final String action, final String type) throws IllegalArgumentException {
        this(subject, resource, action);
        if (type != null) {
            getResourceAttributes().putAttribute(RESOURCE_TYPE, type);
        }
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the action.
     *
     * @return String.
     * @deprecated use getActionId to get the id of the action.
     */
    public String getAction() {
        return getActionId();
    }

    /**
     * Method gets the action attributes.
     *
     * @return Attributes.
     */
    public Attributes getActionAttributes() {
        return m_actionAttributes;
    }

    /**
     * Method gets the actionId.
     *
     * @return String.
     */
    public String getActionId() {
        return getActionAttributes().getStringAttributeValue(ACTION_ID);
    }

    /**
     * Method sets the actionId.
     *
     * @param actionId the action id.
     */
    public void setActionId(final String actionId) {
        getActionAttributes().putAttribute(ACTION_ID, actionId);
    }

    /**
     * Method gets the resource.
     *
     * @return String.
     * @deprecated use getResourceId() to get the id of the resource
     */
    public String getResource() {
        return getResourceId();
    }

    /**
     * Method gets the resource attributes.
     *
     * @return Attributes.
     */
    public Attributes getResourceAttributes() {
        return m_resourceAttributes;
    }

    /**
     * Method gets the resourceId.
     *
     * @return String.
     */
    public String getResourceId() {
        return getResourceAttributes().getStringAttributeValue(RESOURCE_ID);
    }

    /**
     * Method sets the resourceId.
     *
     * @param resourceId the resourceId.
     */
    public void setResourceId(final String resourceId) {
        getResourceAttributes().putAttribute(RESOURCE_ID, resourceId);
    }

    /**
     * Method gets the subject.
     *
     * @return Subject.
     */
    public Subject getSubject() {
        return m_subject;
    }

    /**
     * Method gets the type.
     *
     * @return String.
     * @deprecated use a application specific type attribute.
     */
    public String getType() {
        return getResourceAttributes().getStringAttributeValue(RESOURCE_TYPE);
    }

    /**
     * Method equals.
     *
     * @param o the o.
     * @return boolean.
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof Target)) {
            return false;
        }

        final Target target = (Target) o;

        if (!m_actionAttributes.equals(target.m_actionAttributes)) {
            return false;
        }
        if (!m_resourceAttributes.equals(target.m_resourceAttributes)) {
            return false;
        }
        if (!m_subject.equals(target.m_subject)) {
            return false;
        }

        return true;
    }

    /**
     * Method hashCode.
     *
     * @return int.
     */
    public int hashCode() {
        int result;
        result = m_subject.hashCode();
        result = 31 * result + m_resourceAttributes.hashCode();
        result = 31 * result + m_actionAttributes.hashCode();
        return result;
    }


    /**
     * Method toString.
     *
     * @return String.
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
        sb.append("org.n52.security.decision.Target");
        sb.append("{subject=").append(getSubject());
        sb.append(", resourceAttributes=").append(getResourceAttributes());
        sb.append(", actionAttributes=").append(getActionAttributes());
        sb.append('}');
        return sb.toString();
    }
}